sqlglot.dialects.duckdb API documentation

Edit on GitHub

# sqlglot.dialects.duckdb

View Source

      1from __future__ import annotations
      2
      3import typing as t
      4
      5from sqlglot import exp, generator, parser, tokens, transforms
      6from sqlglot.dialects.dialect import (
      7    Dialect,
      8    NormalizationStrategy,
      9    approx_count_distinct_sql,
     10    arg_max_or_min_no_count,
     11    arrow_json_extract_scalar_sql,
     12    arrow_json_extract_sql,
     13    binary_from_function,
     14    bool_xor_sql,
     15    date_trunc_to_time,
     16    datestrtodate_sql,
     17    encode_decode_sql,
     18    format_time_lambda,
     19    inline_array_sql,
     20    no_comment_column_constraint_sql,
     21    no_properties_sql,
     22    no_safe_divide_sql,
     23    no_timestamp_sql,
     24    pivot_column_names,
     25    regexp_extract_sql,
     26    rename_func,
     27    str_position_sql,
     28    str_to_time_sql,
     29    timestamptrunc_sql,
     30    timestrtotime_sql,
     31    ts_or_ds_to_date_sql,
     32)
     33from sqlglot.helper import seq_get
     34from sqlglot.tokens import TokenType
     35
     36
     37def _ts_or_ds_add_sql(self: DuckDB.Generator, expression: exp.TsOrDsAdd) -> str:
     38    this = self.sql(expression, "this")
     39    unit = self.sql(expression, "unit").strip("'") or "DAY"
     40    interval = self.sql(exp.Interval(this=expression.expression, unit=unit))
     41    return f"CAST({this} AS {self.sql(expression.return_type)}) + {interval}"
     42
     43
     44def _date_delta_sql(self: DuckDB.Generator, expression: exp.DateAdd | exp.DateSub) -> str:
     45    this = self.sql(expression, "this")
     46    unit = self.sql(expression, "unit").strip("'") or "DAY"
     47    op = "+" if isinstance(expression, exp.DateAdd) else "-"
     48    return f"{this} {op} {self.sql(exp.Interval(this=expression.expression, unit=unit))}"
     49
     50
     51# BigQuery -> DuckDB conversion for the DATE function
     52def _date_sql(self: DuckDB.Generator, expression: exp.Date) -> str:
     53    result = f"CAST({self.sql(expression, 'this')} AS DATE)"
     54    zone = self.sql(expression, "zone")
     55
     56    if zone:
     57        date_str = self.func("STRFTIME", result, "'%d/%m/%Y'")
     58        date_str = f"{date_str} || ' ' || {zone}"
     59
     60        # This will create a TIMESTAMP with time zone information
     61        result = self.func("STRPTIME", date_str, "'%d/%m/%Y %Z'")
     62
     63    return result
     64
     65
     66def _array_sort_sql(self: DuckDB.Generator, expression: exp.ArraySort) -> str:
     67    if expression.expression:
     68        self.unsupported("DUCKDB ARRAY_SORT does not support a comparator")
     69    return f"ARRAY_SORT({self.sql(expression, 'this')})"
     70
     71
     72def _sort_array_sql(self: DuckDB.Generator, expression: exp.SortArray) -> str:
     73    this = self.sql(expression, "this")
     74    if expression.args.get("asc") == exp.false():
     75        return f"ARRAY_REVERSE_SORT({this})"
     76    return f"ARRAY_SORT({this})"
     77
     78
     79def _sort_array_reverse(args: t.List) -> exp.Expression:
     80    return exp.SortArray(this=seq_get(args, 0), asc=exp.false())
     81
     82
     83def _parse_date_diff(args: t.List) -> exp.Expression:
     84    return exp.DateDiff(this=seq_get(args, 2), expression=seq_get(args, 1), unit=seq_get(args, 0))
     85
     86
     87def _parse_make_timestamp(args: t.List) -> exp.Expression:
     88    if len(args) == 1:
     89        return exp.UnixToTime(this=seq_get(args, 0), scale=exp.UnixToTime.MICROS)
     90
     91    return exp.TimestampFromParts(
     92        year=seq_get(args, 0),
     93        month=seq_get(args, 1),
     94        day=seq_get(args, 2),
     95        hour=seq_get(args, 3),
     96        min=seq_get(args, 4),
     97        sec=seq_get(args, 5),
     98    )
     99
    100
    101def _struct_sql(self: DuckDB.Generator, expression: exp.Struct) -> str:
    102    args: t.List[str] = []
    103    for expr in expression.expressions:
    104        if isinstance(expr, exp.Alias):
    105            key = expr.alias
    106            value = expr.this
    107        else:
    108            key = expr.name or expr.this.name
    109            if isinstance(expr, exp.Bracket):
    110                value = expr.expressions[0]
    111            else:
    112                value = expr.expression
    113
    114        args.append(f"{self.sql(exp.Literal.string(key))}: {self.sql(value)}")
    115
    116    return f"{{{', '.join(args)}}}"
    117
    118
    119def _datatype_sql(self: DuckDB.Generator, expression: exp.DataType) -> str:
    120    if expression.is_type("array"):
    121        return f"{self.expressions(expression, flat=True)}[]"
    122
    123    # Type TIMESTAMP / TIME WITH TIME ZONE does not support any modifiers
    124    if expression.is_type("timestamptz", "timetz"):
    125        return expression.this.value
    126
    127    return self.datatype_sql(expression)
    128
    129
    130def _json_format_sql(self: DuckDB.Generator, expression: exp.JSONFormat) -> str:
    131    sql = self.func("TO_JSON", expression.this, expression.args.get("options"))
    132    return f"CAST({sql} AS TEXT)"
    133
    134
    135def _unix_to_time_sql(self: DuckDB.Generator, expression: exp.UnixToTime) -> str:
    136    scale = expression.args.get("scale")
    137    timestamp = self.sql(expression, "this")
    138    if scale in (None, exp.UnixToTime.SECONDS):
    139        return f"TO_TIMESTAMP({timestamp})"
    140    if scale == exp.UnixToTime.MILLIS:
    141        return f"EPOCH_MS({timestamp})"
    142    if scale == exp.UnixToTime.MICROS:
    143        return f"MAKE_TIMESTAMP({timestamp})"
    144    if scale == exp.UnixToTime.NANOS:
    145        return f"TO_TIMESTAMP({timestamp} / 1000000000)"
    146
    147    self.unsupported(f"Unsupported scale for timestamp: {scale}.")
    148    return ""
    149
    150
    151class DuckDB(Dialect):
    152    NULL_ORDERING = "nulls_are_last"
    153    SUPPORTS_USER_DEFINED_TYPES = False
    154    SAFE_DIVISION = True
    155    INDEX_OFFSET = 1
    156    CONCAT_COALESCE = True
    157
    158    # https://duckdb.org/docs/sql/introduction.html#creating-a-new-table
    159    NORMALIZATION_STRATEGY = NormalizationStrategy.CASE_INSENSITIVE
    160
    161    class Tokenizer(tokens.Tokenizer):
    162        KEYWORDS = {
    163            **tokens.Tokenizer.KEYWORDS,
    164            "//": TokenType.DIV,
    165            "ATTACH": TokenType.COMMAND,
    166            "BINARY": TokenType.VARBINARY,
    167            "BITSTRING": TokenType.BIT,
    168            "BPCHAR": TokenType.TEXT,
    169            "CHAR": TokenType.TEXT,
    170            "CHARACTER VARYING": TokenType.TEXT,
    171            "EXCLUDE": TokenType.EXCEPT,
    172            "LOGICAL": TokenType.BOOLEAN,
    173            "PIVOT_WIDER": TokenType.PIVOT,
    174            "SIGNED": TokenType.INT,
    175            "STRING": TokenType.VARCHAR,
    176            "UBIGINT": TokenType.UBIGINT,
    177            "UINTEGER": TokenType.UINT,
    178            "USMALLINT": TokenType.USMALLINT,
    179            "UTINYINT": TokenType.UTINYINT,
    180            "TIMESTAMP_S": TokenType.TIMESTAMP_S,
    181            "TIMESTAMP_MS": TokenType.TIMESTAMP_MS,
    182            "TIMESTAMP_NS": TokenType.TIMESTAMP_NS,
    183            "TIMESTAMP_US": TokenType.TIMESTAMP,
    184        }
    185
    186    class Parser(parser.Parser):
    187        BITWISE = {
    188            **parser.Parser.BITWISE,
    189            TokenType.TILDA: exp.RegexpLike,
    190        }
    191
    192        FUNCTIONS = {
    193            **parser.Parser.FUNCTIONS,
    194            "ARRAY_HAS": exp.ArrayContains.from_arg_list,
    195            "ARRAY_LENGTH": exp.ArraySize.from_arg_list,
    196            "ARRAY_SORT": exp.SortArray.from_arg_list,
    197            "ARRAY_REVERSE_SORT": _sort_array_reverse,
    198            "DATEDIFF": _parse_date_diff,
    199            "DATE_DIFF": _parse_date_diff,
    200            "DATE_TRUNC": date_trunc_to_time,
    201            "DATETRUNC": date_trunc_to_time,
    202            "DECODE": lambda args: exp.Decode(
    203                this=seq_get(args, 0), charset=exp.Literal.string("utf-8")
    204            ),
    205            "ENCODE": lambda args: exp.Encode(
    206                this=seq_get(args, 0), charset=exp.Literal.string("utf-8")
    207            ),
    208            "EPOCH": exp.TimeToUnix.from_arg_list,
    209            "EPOCH_MS": lambda args: exp.UnixToTime(
    210                this=seq_get(args, 0), scale=exp.UnixToTime.MILLIS
    211            ),
    212            "LIST_HAS": exp.ArrayContains.from_arg_list,
    213            "LIST_REVERSE_SORT": _sort_array_reverse,
    214            "LIST_SORT": exp.SortArray.from_arg_list,
    215            "LIST_VALUE": exp.Array.from_arg_list,
    216            "MAKE_TIMESTAMP": _parse_make_timestamp,
    217            "MEDIAN": lambda args: exp.PercentileCont(
    218                this=seq_get(args, 0), expression=exp.Literal.number(0.5)
    219            ),
    220            "QUANTILE_CONT": exp.PercentileCont.from_arg_list,
    221            "QUANTILE_DISC": exp.PercentileDisc.from_arg_list,
    222            "REGEXP_EXTRACT": lambda args: exp.RegexpExtract(
    223                this=seq_get(args, 0), expression=seq_get(args, 1), group=seq_get(args, 2)
    224            ),
    225            "REGEXP_MATCHES": exp.RegexpLike.from_arg_list,
    226            "REGEXP_REPLACE": lambda args: exp.RegexpReplace(
    227                this=seq_get(args, 0),
    228                expression=seq_get(args, 1),
    229                replacement=seq_get(args, 2),
    230                modifiers=seq_get(args, 3),
    231            ),
    232            "STRFTIME": format_time_lambda(exp.TimeToStr, "duckdb"),
    233            "STRING_SPLIT": exp.Split.from_arg_list,
    234            "STRING_SPLIT_REGEX": exp.RegexpSplit.from_arg_list,
    235            "STRING_TO_ARRAY": exp.Split.from_arg_list,
    236            "STRPTIME": format_time_lambda(exp.StrToTime, "duckdb"),
    237            "STRUCT_PACK": exp.Struct.from_arg_list,
    238            "STR_SPLIT": exp.Split.from_arg_list,
    239            "STR_SPLIT_REGEX": exp.RegexpSplit.from_arg_list,
    240            "TO_TIMESTAMP": exp.UnixToTime.from_arg_list,
    241            "UNNEST": exp.Explode.from_arg_list,
    242            "XOR": binary_from_function(exp.BitwiseXor),
    243        }
    244
    245        FUNCTION_PARSERS = parser.Parser.FUNCTION_PARSERS.copy()
    246        FUNCTION_PARSERS.pop("DECODE", None)
    247
    248        TABLE_ALIAS_TOKENS = parser.Parser.TABLE_ALIAS_TOKENS - {
    249            TokenType.SEMI,
    250            TokenType.ANTI,
    251        }
    252
    253        def _parse_types(
    254            self, check_func: bool = False, schema: bool = False, allow_identifiers: bool = True
    255        ) -> t.Optional[exp.Expression]:
    256            this = super()._parse_types(
    257                check_func=check_func, schema=schema, allow_identifiers=allow_identifiers
    258            )
    259
    260            # DuckDB treats NUMERIC and DECIMAL without precision as DECIMAL(18, 3)
    261            # See: https://duckdb.org/docs/sql/data_types/numeric
    262            if (
    263                isinstance(this, exp.DataType)
    264                and this.is_type("numeric", "decimal")
    265                and not this.expressions
    266            ):
    267                return exp.DataType.build("DECIMAL(18, 3)")
    268
    269            return this
    270
    271        def _parse_struct_types(self) -> t.Optional[exp.Expression]:
    272            return self._parse_field_def()
    273
    274        def _pivot_column_names(self, aggregations: t.List[exp.Expression]) -> t.List[str]:
    275            if len(aggregations) == 1:
    276                return super()._pivot_column_names(aggregations)
    277            return pivot_column_names(aggregations, dialect="duckdb")
    278
    279    class Generator(generator.Generator):
    280        JOIN_HINTS = False
    281        TABLE_HINTS = False
    282        QUERY_HINTS = False
    283        LIMIT_FETCH = "LIMIT"
    284        STRUCT_DELIMITER = ("(", ")")
    285        RENAME_TABLE_WITH_DB = False
    286        NVL2_SUPPORTED = False
    287        SEMI_ANTI_JOIN_WITH_SIDE = False
    288
    289        TRANSFORMS = {
    290            **generator.Generator.TRANSFORMS,
    291            exp.ApproxDistinct: approx_count_distinct_sql,
    292            exp.Array: lambda self, e: self.func("ARRAY", e.expressions[0])
    293            if e.expressions and e.expressions[0].find(exp.Select)
    294            else inline_array_sql(self, e),
    295            exp.ArraySize: rename_func("ARRAY_LENGTH"),
    296            exp.ArgMax: arg_max_or_min_no_count("ARG_MAX"),
    297            exp.ArgMin: arg_max_or_min_no_count("ARG_MIN"),
    298            exp.ArraySort: _array_sort_sql,
    299            exp.ArraySum: rename_func("LIST_SUM"),
    300            exp.BitwiseXor: rename_func("XOR"),
    301            exp.CommentColumnConstraint: no_comment_column_constraint_sql,
    302            exp.CurrentDate: lambda self, e: "CURRENT_DATE",
    303            exp.CurrentTime: lambda self, e: "CURRENT_TIME",
    304            exp.CurrentTimestamp: lambda self, e: "CURRENT_TIMESTAMP",
    305            exp.DayOfMonth: rename_func("DAYOFMONTH"),
    306            exp.DayOfWeek: rename_func("DAYOFWEEK"),
    307            exp.DayOfYear: rename_func("DAYOFYEAR"),
    308            exp.DataType: _datatype_sql,
    309            exp.Date: _date_sql,
    310            exp.DateAdd: _date_delta_sql,
    311            exp.DateFromParts: rename_func("MAKE_DATE"),
    312            exp.DateSub: _date_delta_sql,
    313            exp.DateDiff: lambda self, e: self.func(
    314                "DATE_DIFF", f"'{e.args.get('unit') or 'day'}'", e.expression, e.this
    315            ),
    316            exp.DateStrToDate: datestrtodate_sql,
    317            exp.DateToDi: lambda self, e: f"CAST(STRFTIME({self.sql(e, 'this')}, {DuckDB.DATEINT_FORMAT}) AS INT)",
    318            exp.Decode: lambda self, e: encode_decode_sql(self, e, "DECODE", replace=False),
    319            exp.DiToDate: lambda self, e: f"CAST(STRPTIME(CAST({self.sql(e, 'this')} AS TEXT), {DuckDB.DATEINT_FORMAT}) AS DATE)",
    320            exp.Encode: lambda self, e: encode_decode_sql(self, e, "ENCODE", replace=False),
    321            exp.Explode: rename_func("UNNEST"),
    322            exp.IntDiv: lambda self, e: self.binary(e, "//"),
    323            exp.IsInf: rename_func("ISINF"),
    324            exp.IsNan: rename_func("ISNAN"),
    325            exp.JSONExtract: arrow_json_extract_sql,
    326            exp.JSONExtractScalar: arrow_json_extract_scalar_sql,
    327            exp.JSONFormat: _json_format_sql,
    328            exp.JSONBExtract: arrow_json_extract_sql,
    329            exp.JSONBExtractScalar: arrow_json_extract_scalar_sql,
    330            exp.LogicalOr: rename_func("BOOL_OR"),
    331            exp.LogicalAnd: rename_func("BOOL_AND"),
    332            exp.MonthsBetween: lambda self, e: self.func(
    333                "DATEDIFF",
    334                "'month'",
    335                exp.cast(e.expression, "timestamp", copy=True),
    336                exp.cast(e.this, "timestamp", copy=True),
    337            ),
    338            exp.ParseJSON: rename_func("JSON"),
    339            exp.PercentileCont: rename_func("QUANTILE_CONT"),
    340            exp.PercentileDisc: rename_func("QUANTILE_DISC"),
    341            # DuckDB doesn't allow qualified columns inside of PIVOT expressions.
    342            # See: https://github.com/duckdb/duckdb/blob/671faf92411182f81dce42ac43de8bfb05d9909e/src/planner/binder/tableref/bind_pivot.cpp#L61-L62
    343            exp.Pivot: transforms.preprocess([transforms.unqualify_columns]),
    344            exp.Properties: no_properties_sql,
    345            exp.RegexpExtract: regexp_extract_sql,
    346            exp.RegexpReplace: lambda self, e: self.func(
    347                "REGEXP_REPLACE",
    348                e.this,
    349                e.expression,
    350                e.args.get("replacement"),
    351                e.args.get("modifiers"),
    352            ),
    353            exp.RegexpLike: rename_func("REGEXP_MATCHES"),
    354            exp.RegexpSplit: rename_func("STR_SPLIT_REGEX"),
    355            exp.SafeDivide: no_safe_divide_sql,
    356            exp.Split: rename_func("STR_SPLIT"),
    357            exp.SortArray: _sort_array_sql,
    358            exp.StrPosition: str_position_sql,
    359            exp.StrToDate: lambda self, e: f"CAST({str_to_time_sql(self, e)} AS DATE)",
    360            exp.StrToTime: str_to_time_sql,
    361            exp.StrToUnix: lambda self, e: f"EPOCH(STRPTIME({self.sql(e, 'this')}, {self.format_time(e)}))",
    362            exp.Struct: _struct_sql,
    363            exp.Timestamp: no_timestamp_sql,
    364            exp.TimestampFromParts: rename_func("MAKE_TIMESTAMP"),
    365            exp.TimestampTrunc: timestamptrunc_sql,
    366            exp.TimeStrToDate: lambda self, e: f"CAST({self.sql(e, 'this')} AS DATE)",
    367            exp.TimeStrToTime: timestrtotime_sql,
    368            exp.TimeStrToUnix: lambda self, e: f"EPOCH(CAST({self.sql(e, 'this')} AS TIMESTAMP))",
    369            exp.TimeToStr: lambda self, e: f"STRFTIME({self.sql(e, 'this')}, {self.format_time(e)})",
    370            exp.TimeToUnix: rename_func("EPOCH"),
    371            exp.TsOrDiToDi: lambda self, e: f"CAST(SUBSTR(REPLACE(CAST({self.sql(e, 'this')} AS TEXT), '-', ''), 1, 8) AS INT)",
    372            exp.TsOrDsAdd: _ts_or_ds_add_sql,
    373            exp.TsOrDsDiff: lambda self, e: self.func(
    374                "DATE_DIFF",
    375                f"'{e.args.get('unit') or 'day'}'",
    376                exp.cast(e.expression, "TIMESTAMP"),
    377                exp.cast(e.this, "TIMESTAMP"),
    378            ),
    379            exp.TsOrDsToDate: ts_or_ds_to_date_sql("duckdb"),
    380            exp.UnixToStr: lambda self, e: f"STRFTIME(TO_TIMESTAMP({self.sql(e, 'this')}), {self.format_time(e)})",
    381            exp.UnixToTime: _unix_to_time_sql,
    382            exp.UnixToTimeStr: lambda self, e: f"CAST(TO_TIMESTAMP({self.sql(e, 'this')}) AS TEXT)",
    383            exp.VariancePop: rename_func("VAR_POP"),
    384            exp.WeekOfYear: rename_func("WEEKOFYEAR"),
    385            exp.Xor: bool_xor_sql,
    386        }
    387
    388        TYPE_MAPPING = {
    389            **generator.Generator.TYPE_MAPPING,
    390            exp.DataType.Type.BINARY: "BLOB",
    391            exp.DataType.Type.CHAR: "TEXT",
    392            exp.DataType.Type.FLOAT: "REAL",
    393            exp.DataType.Type.NCHAR: "TEXT",
    394            exp.DataType.Type.NVARCHAR: "TEXT",
    395            exp.DataType.Type.UINT: "UINTEGER",
    396            exp.DataType.Type.VARBINARY: "BLOB",
    397            exp.DataType.Type.VARCHAR: "TEXT",
    398            exp.DataType.Type.TIMESTAMP_S: "TIMESTAMP_S",
    399            exp.DataType.Type.TIMESTAMP_MS: "TIMESTAMP_MS",
    400            exp.DataType.Type.TIMESTAMP_NS: "TIMESTAMP_NS",
    401        }
    402
    403        STAR_MAPPING = {**generator.Generator.STAR_MAPPING, "except": "EXCLUDE"}
    404
    405        UNWRAPPED_INTERVAL_VALUES = (exp.Column, exp.Literal, exp.Paren)
    406
    407        PROPERTIES_LOCATION = {
    408            **generator.Generator.PROPERTIES_LOCATION,
    409            exp.VolatileProperty: exp.Properties.Location.UNSUPPORTED,
    410        }
    411
    412        def interval_sql(self, expression: exp.Interval) -> str:
    413            multiplier: t.Optional[int] = None
    414            unit = expression.text("unit").lower()
    415
    416            if unit.startswith("week"):
    417                multiplier = 7
    418            if unit.startswith("quarter"):
    419                multiplier = 90
    420
    421            if multiplier:
    422                return f"({multiplier} * {super().interval_sql(exp.Interval(this=expression.this, unit=exp.var('day')))})"
    423
    424            return super().interval_sql(expression)
    425
    426        def tablesample_sql(
    427            self, expression: exp.TableSample, seed_prefix: str = "SEED", sep: str = " AS "
    428        ) -> str:
    429            return super().tablesample_sql(expression, seed_prefix="REPEATABLE", sep=sep)

class DuckDB(sqlglot.dialects.dialect.Dialect): View Source

    152class DuckDB(Dialect):
    153    NULL_ORDERING = "nulls_are_last"
    154    SUPPORTS_USER_DEFINED_TYPES = False
    155    SAFE_DIVISION = True
    156    INDEX_OFFSET = 1
    157    CONCAT_COALESCE = True
    158
    159    # https://duckdb.org/docs/sql/introduction.html#creating-a-new-table
    160    NORMALIZATION_STRATEGY = NormalizationStrategy.CASE_INSENSITIVE
    161
    162    class Tokenizer(tokens.Tokenizer):
    163        KEYWORDS = {
    164            **tokens.Tokenizer.KEYWORDS,
    165            "//": TokenType.DIV,
    166            "ATTACH": TokenType.COMMAND,
    167            "BINARY": TokenType.VARBINARY,
    168            "BITSTRING": TokenType.BIT,
    169            "BPCHAR": TokenType.TEXT,
    170            "CHAR": TokenType.TEXT,
    171            "CHARACTER VARYING": TokenType.TEXT,
    172            "EXCLUDE": TokenType.EXCEPT,
    173            "LOGICAL": TokenType.BOOLEAN,
    174            "PIVOT_WIDER": TokenType.PIVOT,
    175            "SIGNED": TokenType.INT,
    176            "STRING": TokenType.VARCHAR,
    177            "UBIGINT": TokenType.UBIGINT,
    178            "UINTEGER": TokenType.UINT,
    179            "USMALLINT": TokenType.USMALLINT,
    180            "UTINYINT": TokenType.UTINYINT,
    181            "TIMESTAMP_S": TokenType.TIMESTAMP_S,
    182            "TIMESTAMP_MS": TokenType.TIMESTAMP_MS,
    183            "TIMESTAMP_NS": TokenType.TIMESTAMP_NS,
    184            "TIMESTAMP_US": TokenType.TIMESTAMP,
    185        }
    186
    187    class Parser(parser.Parser):
    188        BITWISE = {
    189            **parser.Parser.BITWISE,
    190            TokenType.TILDA: exp.RegexpLike,
    191        }
    192
    193        FUNCTIONS = {
    194            **parser.Parser.FUNCTIONS,
    195            "ARRAY_HAS": exp.ArrayContains.from_arg_list,
    196            "ARRAY_LENGTH": exp.ArraySize.from_arg_list,
    197            "ARRAY_SORT": exp.SortArray.from_arg_list,
    198            "ARRAY_REVERSE_SORT": _sort_array_reverse,
    199            "DATEDIFF": _parse_date_diff,
    200            "DATE_DIFF": _parse_date_diff,
    201            "DATE_TRUNC": date_trunc_to_time,
    202            "DATETRUNC": date_trunc_to_time,
    203            "DECODE": lambda args: exp.Decode(
    204                this=seq_get(args, 0), charset=exp.Literal.string("utf-8")
    205            ),
    206            "ENCODE": lambda args: exp.Encode(
    207                this=seq_get(args, 0), charset=exp.Literal.string("utf-8")
    208            ),
    209            "EPOCH": exp.TimeToUnix.from_arg_list,
    210            "EPOCH_MS": lambda args: exp.UnixToTime(
    211                this=seq_get(args, 0), scale=exp.UnixToTime.MILLIS
    212            ),
    213            "LIST_HAS": exp.ArrayContains.from_arg_list,
    214            "LIST_REVERSE_SORT": _sort_array_reverse,
    215            "LIST_SORT": exp.SortArray.from_arg_list,
    216            "LIST_VALUE": exp.Array.from_arg_list,
    217            "MAKE_TIMESTAMP": _parse_make_timestamp,
    218            "MEDIAN": lambda args: exp.PercentileCont(
    219                this=seq_get(args, 0), expression=exp.Literal.number(0.5)
    220            ),
    221            "QUANTILE_CONT": exp.PercentileCont.from_arg_list,
    222            "QUANTILE_DISC": exp.PercentileDisc.from_arg_list,
    223            "REGEXP_EXTRACT": lambda args: exp.RegexpExtract(
    224                this=seq_get(args, 0), expression=seq_get(args, 1), group=seq_get(args, 2)
    225            ),
    226            "REGEXP_MATCHES": exp.RegexpLike.from_arg_list,
    227            "REGEXP_REPLACE": lambda args: exp.RegexpReplace(
    228                this=seq_get(args, 0),
    229                expression=seq_get(args, 1),
    230                replacement=seq_get(args, 2),
    231                modifiers=seq_get(args, 3),
    232            ),
    233            "STRFTIME": format_time_lambda(exp.TimeToStr, "duckdb"),
    234            "STRING_SPLIT": exp.Split.from_arg_list,
    235            "STRING_SPLIT_REGEX": exp.RegexpSplit.from_arg_list,
    236            "STRING_TO_ARRAY": exp.Split.from_arg_list,
    237            "STRPTIME": format_time_lambda(exp.StrToTime, "duckdb"),
    238            "STRUCT_PACK": exp.Struct.from_arg_list,
    239            "STR_SPLIT": exp.Split.from_arg_list,
    240            "STR_SPLIT_REGEX": exp.RegexpSplit.from_arg_list,
    241            "TO_TIMESTAMP": exp.UnixToTime.from_arg_list,
    242            "UNNEST": exp.Explode.from_arg_list,
    243            "XOR": binary_from_function(exp.BitwiseXor),
    244        }
    245
    246        FUNCTION_PARSERS = parser.Parser.FUNCTION_PARSERS.copy()
    247        FUNCTION_PARSERS.pop("DECODE", None)
    248
    249        TABLE_ALIAS_TOKENS = parser.Parser.TABLE_ALIAS_TOKENS - {
    250            TokenType.SEMI,
    251            TokenType.ANTI,
    252        }
    253
    254        def _parse_types(
    255            self, check_func: bool = False, schema: bool = False, allow_identifiers: bool = True
    256        ) -> t.Optional[exp.Expression]:
    257            this = super()._parse_types(
    258                check_func=check_func, schema=schema, allow_identifiers=allow_identifiers
    259            )
    260
    261            # DuckDB treats NUMERIC and DECIMAL without precision as DECIMAL(18, 3)
    262            # See: https://duckdb.org/docs/sql/data_types/numeric
    263            if (
    264                isinstance(this, exp.DataType)
    265                and this.is_type("numeric", "decimal")
    266                and not this.expressions
    267            ):
    268                return exp.DataType.build("DECIMAL(18, 3)")
    269
    270            return this
    271
    272        def _parse_struct_types(self) -> t.Optional[exp.Expression]:
    273            return self._parse_field_def()
    274
    275        def _pivot_column_names(self, aggregations: t.List[exp.Expression]) -> t.List[str]:
    276            if len(aggregations) == 1:
    277                return super()._pivot_column_names(aggregations)
    278            return pivot_column_names(aggregations, dialect="duckdb")
    279
    280    class Generator(generator.Generator):
    281        JOIN_HINTS = False
    282        TABLE_HINTS = False
    283        QUERY_HINTS = False
    284        LIMIT_FETCH = "LIMIT"
    285        STRUCT_DELIMITER = ("(", ")")
    286        RENAME_TABLE_WITH_DB = False
    287        NVL2_SUPPORTED = False
    288        SEMI_ANTI_JOIN_WITH_SIDE = False
    289
    290        TRANSFORMS = {
    291            **generator.Generator.TRANSFORMS,
    292            exp.ApproxDistinct: approx_count_distinct_sql,
    293            exp.Array: lambda self, e: self.func("ARRAY", e.expressions[0])
    294            if e.expressions and e.expressions[0].find(exp.Select)
    295            else inline_array_sql(self, e),
    296            exp.ArraySize: rename_func("ARRAY_LENGTH"),
    297            exp.ArgMax: arg_max_or_min_no_count("ARG_MAX"),
    298            exp.ArgMin: arg_max_or_min_no_count("ARG_MIN"),
    299            exp.ArraySort: _array_sort_sql,
    300            exp.ArraySum: rename_func("LIST_SUM"),
    301            exp.BitwiseXor: rename_func("XOR"),
    302            exp.CommentColumnConstraint: no_comment_column_constraint_sql,
    303            exp.CurrentDate: lambda self, e: "CURRENT_DATE",
    304            exp.CurrentTime: lambda self, e: "CURRENT_TIME",
    305            exp.CurrentTimestamp: lambda self, e: "CURRENT_TIMESTAMP",
    306            exp.DayOfMonth: rename_func("DAYOFMONTH"),
    307            exp.DayOfWeek: rename_func("DAYOFWEEK"),
    308            exp.DayOfYear: rename_func("DAYOFYEAR"),
    309            exp.DataType: _datatype_sql,
    310            exp.Date: _date_sql,
    311            exp.DateAdd: _date_delta_sql,
    312            exp.DateFromParts: rename_func("MAKE_DATE"),
    313            exp.DateSub: _date_delta_sql,
    314            exp.DateDiff: lambda self, e: self.func(
    315                "DATE_DIFF", f"'{e.args.get('unit') or 'day'}'", e.expression, e.this
    316            ),
    317            exp.DateStrToDate: datestrtodate_sql,
    318            exp.DateToDi: lambda self, e: f"CAST(STRFTIME({self.sql(e, 'this')}, {DuckDB.DATEINT_FORMAT}) AS INT)",
    319            exp.Decode: lambda self, e: encode_decode_sql(self, e, "DECODE", replace=False),
    320            exp.DiToDate: lambda self, e: f"CAST(STRPTIME(CAST({self.sql(e, 'this')} AS TEXT), {DuckDB.DATEINT_FORMAT}) AS DATE)",
    321            exp.Encode: lambda self, e: encode_decode_sql(self, e, "ENCODE", replace=False),
    322            exp.Explode: rename_func("UNNEST"),
    323            exp.IntDiv: lambda self, e: self.binary(e, "//"),
    324            exp.IsInf: rename_func("ISINF"),
    325            exp.IsNan: rename_func("ISNAN"),
    326            exp.JSONExtract: arrow_json_extract_sql,
    327            exp.JSONExtractScalar: arrow_json_extract_scalar_sql,
    328            exp.JSONFormat: _json_format_sql,
    329            exp.JSONBExtract: arrow_json_extract_sql,
    330            exp.JSONBExtractScalar: arrow_json_extract_scalar_sql,
    331            exp.LogicalOr: rename_func("BOOL_OR"),
    332            exp.LogicalAnd: rename_func("BOOL_AND"),
    333            exp.MonthsBetween: lambda self, e: self.func(
    334                "DATEDIFF",
    335                "'month'",
    336                exp.cast(e.expression, "timestamp", copy=True),
    337                exp.cast(e.this, "timestamp", copy=True),
    338            ),
    339            exp.ParseJSON: rename_func("JSON"),
    340            exp.PercentileCont: rename_func("QUANTILE_CONT"),
    341            exp.PercentileDisc: rename_func("QUANTILE_DISC"),
    342            # DuckDB doesn't allow qualified columns inside of PIVOT expressions.
    343            # See: https://github.com/duckdb/duckdb/blob/671faf92411182f81dce42ac43de8bfb05d9909e/src/planner/binder/tableref/bind_pivot.cpp#L61-L62
    344            exp.Pivot: transforms.preprocess([transforms.unqualify_columns]),
    345            exp.Properties: no_properties_sql,
    346            exp.RegexpExtract: regexp_extract_sql,
    347            exp.RegexpReplace: lambda self, e: self.func(
    348                "REGEXP_REPLACE",
    349                e.this,
    350                e.expression,
    351                e.args.get("replacement"),
    352                e.args.get("modifiers"),
    353            ),
    354            exp.RegexpLike: rename_func("REGEXP_MATCHES"),
    355            exp.RegexpSplit: rename_func("STR_SPLIT_REGEX"),
    356            exp.SafeDivide: no_safe_divide_sql,
    357            exp.Split: rename_func("STR_SPLIT"),
    358            exp.SortArray: _sort_array_sql,
    359            exp.StrPosition: str_position_sql,
    360            exp.StrToDate: lambda self, e: f"CAST({str_to_time_sql(self, e)} AS DATE)",
    361            exp.StrToTime: str_to_time_sql,
    362            exp.StrToUnix: lambda self, e: f"EPOCH(STRPTIME({self.sql(e, 'this')}, {self.format_time(e)}))",
    363            exp.Struct: _struct_sql,
    364            exp.Timestamp: no_timestamp_sql,
    365            exp.TimestampFromParts: rename_func("MAKE_TIMESTAMP"),
    366            exp.TimestampTrunc: timestamptrunc_sql,
    367            exp.TimeStrToDate: lambda self, e: f"CAST({self.sql(e, 'this')} AS DATE)",
    368            exp.TimeStrToTime: timestrtotime_sql,
    369            exp.TimeStrToUnix: lambda self, e: f"EPOCH(CAST({self.sql(e, 'this')} AS TIMESTAMP))",
    370            exp.TimeToStr: lambda self, e: f"STRFTIME({self.sql(e, 'this')}, {self.format_time(e)})",
    371            exp.TimeToUnix: rename_func("EPOCH"),
    372            exp.TsOrDiToDi: lambda self, e: f"CAST(SUBSTR(REPLACE(CAST({self.sql(e, 'this')} AS TEXT), '-', ''), 1, 8) AS INT)",
    373            exp.TsOrDsAdd: _ts_or_ds_add_sql,
    374            exp.TsOrDsDiff: lambda self, e: self.func(
    375                "DATE_DIFF",
    376                f"'{e.args.get('unit') or 'day'}'",
    377                exp.cast(e.expression, "TIMESTAMP"),
    378                exp.cast(e.this, "TIMESTAMP"),
    379            ),
    380            exp.TsOrDsToDate: ts_or_ds_to_date_sql("duckdb"),
    381            exp.UnixToStr: lambda self, e: f"STRFTIME(TO_TIMESTAMP({self.sql(e, 'this')}), {self.format_time(e)})",
    382            exp.UnixToTime: _unix_to_time_sql,
    383            exp.UnixToTimeStr: lambda self, e: f"CAST(TO_TIMESTAMP({self.sql(e, 'this')}) AS TEXT)",
    384            exp.VariancePop: rename_func("VAR_POP"),
    385            exp.WeekOfYear: rename_func("WEEKOFYEAR"),
    386            exp.Xor: bool_xor_sql,
    387        }
    388
    389        TYPE_MAPPING = {
    390            **generator.Generator.TYPE_MAPPING,
    391            exp.DataType.Type.BINARY: "BLOB",
    392            exp.DataType.Type.CHAR: "TEXT",
    393            exp.DataType.Type.FLOAT: "REAL",
    394            exp.DataType.Type.NCHAR: "TEXT",
    395            exp.DataType.Type.NVARCHAR: "TEXT",
    396            exp.DataType.Type.UINT: "UINTEGER",
    397            exp.DataType.Type.VARBINARY: "BLOB",
    398            exp.DataType.Type.VARCHAR: "TEXT",
    399            exp.DataType.Type.TIMESTAMP_S: "TIMESTAMP_S",
    400            exp.DataType.Type.TIMESTAMP_MS: "TIMESTAMP_MS",
    401            exp.DataType.Type.TIMESTAMP_NS: "TIMESTAMP_NS",
    402        }
    403
    404        STAR_MAPPING = {**generator.Generator.STAR_MAPPING, "except": "EXCLUDE"}
    405
    406        UNWRAPPED_INTERVAL_VALUES = (exp.Column, exp.Literal, exp.Paren)
    407
    408        PROPERTIES_LOCATION = {
    409            **generator.Generator.PROPERTIES_LOCATION,
    410            exp.VolatileProperty: exp.Properties.Location.UNSUPPORTED,
    411        }
    412
    413        def interval_sql(self, expression: exp.Interval) -> str:
    414            multiplier: t.Optional[int] = None
    415            unit = expression.text("unit").lower()
    416
    417            if unit.startswith("week"):
    418                multiplier = 7
    419            if unit.startswith("quarter"):
    420                multiplier = 90
    421
    422            if multiplier:
    423                return f"({multiplier} * {super().interval_sql(exp.Interval(this=expression.this, unit=exp.var('day')))})"
    424
    425            return super().interval_sql(expression)
    426
    427        def tablesample_sql(
    428            self, expression: exp.TableSample, seed_prefix: str = "SEED", sep: str = " AS "
    429        ) -> str:
    430            return super().tablesample_sql(expression, seed_prefix="REPEATABLE", sep=sep)

NULL_ORDERING = 'nulls_are_last'
SUPPORTS_USER_DEFINED_TYPES = False
SAFE_DIVISION = True
INDEX_OFFSET = 1
CONCAT_COALESCE = True
NORMALIZATION_STRATEGY = <NormalizationStrategy.CASE_INSENSITIVE: 'CASE_INSENSITIVE'>
tokenizer_class = <class 'DuckDB.Tokenizer'>
parser_class = <class 'DuckDB.Parser'>
generator_class = <class 'DuckDB.Generator'>
TIME_TRIE: Dict = {}
FORMAT_TRIE: Dict = {}
INVERSE_TIME_MAPPING: Dict[str, str] = {}
INVERSE_TIME_TRIE: Dict = {}
INVERSE_ESCAPE_SEQUENCES: Dict[str, str] = {}
QUOTE_START = "'"
QUOTE_END = "'"
IDENTIFIER_START = '"'
IDENTIFIER_END = '"'
BIT_START: Optional[str] = None
BIT_END: Optional[str] = None
HEX_START: Optional[str] = None
HEX_END: Optional[str] = None
BYTE_START: Optional[str] = None
BYTE_END: Optional[str] = None

##### Inherited Members

sqlglot.dialects.dialect.Dialect
  Dialect
  UNNEST_COLUMN_ONLY
  ALIAS_POST_TABLESAMPLE
  IDENTIFIERS_CAN_START_WITH_DIGIT
  DPIPE_IS_STRING_CONCAT
  STRICT_STRING_CONCAT
  SUPPORTS_SEMI_ANTI_JOIN
  NORMALIZE_FUNCTIONS
  LOG_BASE_FIRST
  TYPED_DIVISION
  DATE_FORMAT
  DATEINT_FORMAT
  TIME_FORMAT
  TIME_MAPPING
  FORMAT_MAPPING
  ESCAPE_SEQUENCES
  PSEUDOCOLUMNS
  get_or_raise
  format_time
  normalize_identifier
  case_sensitive
  can_identify
  quote_identifier
  parse
  parse_into
  generate
  transpile
  tokenize
  tokenizer
  parser
  generator

class DuckDB.Tokenizer(sqlglot.tokens.Tokenizer): View Source

    162    class Tokenizer(tokens.Tokenizer):
    163        KEYWORDS = {
    164            **tokens.Tokenizer.KEYWORDS,
    165            "//": TokenType.DIV,
    166            "ATTACH": TokenType.COMMAND,
    167            "BINARY": TokenType.VARBINARY,
    168            "BITSTRING": TokenType.BIT,
    169            "BPCHAR": TokenType.TEXT,
    170            "CHAR": TokenType.TEXT,
    171            "CHARACTER VARYING": TokenType.TEXT,
    172            "EXCLUDE": TokenType.EXCEPT,
    173            "LOGICAL": TokenType.BOOLEAN,
    174            "PIVOT_WIDER": TokenType.PIVOT,
    175            "SIGNED": TokenType.INT,
    176            "STRING": TokenType.VARCHAR,
    177            "UBIGINT": TokenType.UBIGINT,
    178            "UINTEGER": TokenType.UINT,
    179            "USMALLINT": TokenType.USMALLINT,
    180            "UTINYINT": TokenType.UTINYINT,
    181            "TIMESTAMP_S": TokenType.TIMESTAMP_S,
    182            "TIMESTAMP_MS": TokenType.TIMESTAMP_MS,
    183            "TIMESTAMP_NS": TokenType.TIMESTAMP_NS,
    184            "TIMESTAMP_US": TokenType.TIMESTAMP,
    185        }

KEYWORDS = {'{%': <TokenType.BLOCK_START: 'BLOCK_START'>, '{%+': <TokenType.BLOCK_START: 'BLOCK_START'>, '{%-': <TokenType.BLOCK_START: 'BLOCK_START'>, '%}': <TokenType.BLOCK_END: 'BLOCK_END'>, '+%}': <TokenType.BLOCK_END: 'BLOCK_END'>, '-%}': <TokenType.BLOCK_END: 'BLOCK_END'>, '{{+': <TokenType.BLOCK_START: 'BLOCK_START'>, '{{-': <TokenType.BLOCK_START: 'BLOCK_START'>, '+}}': <TokenType.BLOCK_END: 'BLOCK_END'>, '-}}': <TokenType.BLOCK_END: 'BLOCK_END'>, '/*+': <TokenType.HINT: 'HINT'>, '==': <TokenType.EQ: 'EQ'>, '::': <TokenType.DCOLON: 'DCOLON'>, '||': <TokenType.DPIPE: 'DPIPE'>, '>=': <TokenType.GTE: 'GTE'>, '<=': <TokenType.LTE: 'LTE'>, '<>': <TokenType.NEQ: 'NEQ'>, '!=': <TokenType.NEQ: 'NEQ'>, ':=': <TokenType.COLON_EQ: 'COLON_EQ'>, '<=>': <TokenType.NULLSAFE_EQ: 'NULLSAFE_EQ'>, '->': <TokenType.ARROW: 'ARROW'>, '->>': <TokenType.DARROW: 'DARROW'>, '=>': <TokenType.FARROW: 'FARROW'>, '#>': <TokenType.HASH_ARROW: 'HASH_ARROW'>, '#>>': <TokenType.DHASH_ARROW: 'DHASH_ARROW'>, '<->': <TokenType.LR_ARROW: 'LR_ARROW'>, '&&': <TokenType.DAMP: 'DAMP'>, '??': <TokenType.DQMARK: 'DQMARK'>, 'ALL': <TokenType.ALL: 'ALL'>, 'ALWAYS': <TokenType.ALWAYS: 'ALWAYS'>, 'AND': <TokenType.AND: 'AND'>, 'ANTI': <TokenType.ANTI: 'ANTI'>, 'ANY': <TokenType.ANY: 'ANY'>, 'ASC': <TokenType.ASC: 'ASC'>, 'AS': <TokenType.ALIAS: 'ALIAS'>, 'ASOF': <TokenType.ASOF: 'ASOF'>, 'AUTOINCREMENT': <TokenType.AUTO_INCREMENT: 'AUTO_INCREMENT'>, 'AUTO_INCREMENT': <TokenType.AUTO_INCREMENT: 'AUTO_INCREMENT'>, 'BEGIN': <TokenType.BEGIN: 'BEGIN'>, 'BETWEEN': <TokenType.BETWEEN: 'BETWEEN'>, 'CACHE': <TokenType.CACHE: 'CACHE'>, 'UNCACHE': <TokenType.UNCACHE: 'UNCACHE'>, 'CASE': <TokenType.CASE: 'CASE'>, 'CHARACTER SET': <TokenType.CHARACTER_SET: 'CHARACTER_SET'>, 'CLUSTER BY': <TokenType.CLUSTER_BY: 'CLUSTER_BY'>, 'COLLATE': <TokenType.COLLATE: 'COLLATE'>, 'COLUMN': <TokenType.COLUMN: 'COLUMN'>, 'COMMIT': <TokenType.COMMIT: 'COMMIT'>, 'CONNECT BY': <TokenType.CONNECT_BY: 'CONNECT_BY'>, 'CONSTRAINT': <TokenType.CONSTRAINT: 'CONSTRAINT'>, 'CREATE': <TokenType.CREATE: 'CREATE'>, 'CROSS': <TokenType.CROSS: 'CROSS'>, 'CUBE': <TokenType.CUBE: 'CUBE'>, 'CURRENT_DATE': <TokenType.CURRENT_DATE: 'CURRENT_DATE'>, 'CURRENT_TIME': <TokenType.CURRENT_TIME: 'CURRENT_TIME'>, 'CURRENT_TIMESTAMP': <TokenType.CURRENT_TIMESTAMP: 'CURRENT_TIMESTAMP'>, 'CURRENT_USER': <TokenType.CURRENT_USER: 'CURRENT_USER'>, 'DATABASE': <TokenType.DATABASE: 'DATABASE'>, 'DEFAULT': <TokenType.DEFAULT: 'DEFAULT'>, 'DELETE': <TokenType.DELETE: 'DELETE'>, 'DESC': <TokenType.DESC: 'DESC'>, 'DESCRIBE': <TokenType.DESCRIBE: 'DESCRIBE'>, 'DISTINCT': <TokenType.DISTINCT: 'DISTINCT'>, 'DISTRIBUTE BY': <TokenType.DISTRIBUTE_BY: 'DISTRIBUTE_BY'>, 'DIV': <TokenType.DIV: 'DIV'>, 'DROP': <TokenType.DROP: 'DROP'>, 'ELSE': <TokenType.ELSE: 'ELSE'>, 'END': <TokenType.END: 'END'>, 'ESCAPE': <TokenType.ESCAPE: 'ESCAPE'>, 'EXCEPT': <TokenType.EXCEPT: 'EXCEPT'>, 'EXECUTE': <TokenType.EXECUTE: 'EXECUTE'>, 'EXISTS': <TokenType.EXISTS: 'EXISTS'>, 'FALSE': <TokenType.FALSE: 'FALSE'>, 'FETCH': <TokenType.FETCH: 'FETCH'>, 'FILTER': <TokenType.FILTER: 'FILTER'>, 'FIRST': <TokenType.FIRST: 'FIRST'>, 'FULL': <TokenType.FULL: 'FULL'>, 'FUNCTION': <TokenType.FUNCTION: 'FUNCTION'>, 'FOR': <TokenType.FOR: 'FOR'>, 'FOREIGN KEY': <TokenType.FOREIGN_KEY: 'FOREIGN_KEY'>, 'FORMAT': <TokenType.FORMAT: 'FORMAT'>, 'FROM': <TokenType.FROM: 'FROM'>, 'GEOGRAPHY': <TokenType.GEOGRAPHY: 'GEOGRAPHY'>, 'GEOMETRY': <TokenType.GEOMETRY: 'GEOMETRY'>, 'GLOB': <TokenType.GLOB: 'GLOB'>, 'GROUP BY': <TokenType.GROUP_BY: 'GROUP_BY'>, 'GROUPING SETS': <TokenType.GROUPING_SETS: 'GROUPING_SETS'>, 'HAVING': <TokenType.HAVING: 'HAVING'>, 'ILIKE': <TokenType.ILIKE: 'ILIKE'>, 'IN': <TokenType.IN: 'IN'>, 'INDEX': <TokenType.INDEX: 'INDEX'>, 'INET': <TokenType.INET: 'INET'>, 'INNER': <TokenType.INNER: 'INNER'>, 'INSERT': <TokenType.INSERT: 'INSERT'>, 'INTERVAL': <TokenType.INTERVAL: 'INTERVAL'>, 'INTERSECT': <TokenType.INTERSECT: 'INTERSECT'>, 'INTO': <TokenType.INTO: 'INTO'>, 'IS': <TokenType.IS: 'IS'>, 'ISNULL': <TokenType.ISNULL: 'ISNULL'>, 'JOIN': <TokenType.JOIN: 'JOIN'>, 'KEEP': <TokenType.KEEP: 'KEEP'>, 'KILL': <TokenType.KILL: 'KILL'>, 'LATERAL': <TokenType.LATERAL: 'LATERAL'>, 'LEFT': <TokenType.LEFT: 'LEFT'>, 'LIKE': <TokenType.LIKE: 'LIKE'>, 'LIMIT': <TokenType.LIMIT: 'LIMIT'>, 'LOAD': <TokenType.LOAD: 'LOAD'>, 'LOCK': <TokenType.LOCK: 'LOCK'>, 'MERGE': <TokenType.MERGE: 'MERGE'>, 'NATURAL': <TokenType.NATURAL: 'NATURAL'>, 'NEXT': <TokenType.NEXT: 'NEXT'>, 'NOT': <TokenType.NOT: 'NOT'>, 'NOTNULL': <TokenType.NOTNULL: 'NOTNULL'>, 'NULL': <TokenType.NULL: 'NULL'>, 'OBJECT': <TokenType.OBJECT: 'OBJECT'>, 'OFFSET': <TokenType.OFFSET: 'OFFSET'>, 'ON': <TokenType.ON: 'ON'>, 'OR': <TokenType.OR: 'OR'>, 'XOR': <TokenType.XOR: 'XOR'>, 'ORDER BY': <TokenType.ORDER_BY: 'ORDER_BY'>, 'ORDINALITY': <TokenType.ORDINALITY: 'ORDINALITY'>, 'OUTER': <TokenType.OUTER: 'OUTER'>, 'OVER': <TokenType.OVER: 'OVER'>, 'OVERLAPS': <TokenType.OVERLAPS: 'OVERLAPS'>, 'OVERWRITE': <TokenType.OVERWRITE: 'OVERWRITE'>, 'PARTITION': <TokenType.PARTITION: 'PARTITION'>, 'PARTITION BY': <TokenType.PARTITION_BY: 'PARTITION_BY'>, 'PARTITIONED BY': <TokenType.PARTITION_BY: 'PARTITION_BY'>, 'PARTITIONED_BY': <TokenType.PARTITION_BY: 'PARTITION_BY'>, 'PERCENT': <TokenType.PERCENT: 'PERCENT'>, 'PIVOT': <TokenType.PIVOT: 'PIVOT'>, 'PRAGMA': <TokenType.PRAGMA: 'PRAGMA'>, 'PRIMARY KEY': <TokenType.PRIMARY_KEY: 'PRIMARY_KEY'>, 'PROCEDURE': <TokenType.PROCEDURE: 'PROCEDURE'>, 'QUALIFY': <TokenType.QUALIFY: 'QUALIFY'>, 'RANGE': <TokenType.RANGE: 'RANGE'>, 'RECURSIVE': <TokenType.RECURSIVE: 'RECURSIVE'>, 'REGEXP': <TokenType.RLIKE: 'RLIKE'>, 'REPLACE': <TokenType.REPLACE: 'REPLACE'>, 'RETURNING': <TokenType.RETURNING: 'RETURNING'>, 'REFERENCES': <TokenType.REFERENCES: 'REFERENCES'>, 'RIGHT': <TokenType.RIGHT: 'RIGHT'>, 'RLIKE': <TokenType.RLIKE: 'RLIKE'>, 'ROLLBACK': <TokenType.ROLLBACK: 'ROLLBACK'>, 'ROLLUP': <TokenType.ROLLUP: 'ROLLUP'>, 'ROW': <TokenType.ROW: 'ROW'>, 'ROWS': <TokenType.ROWS: 'ROWS'>, 'SCHEMA': <TokenType.SCHEMA: 'SCHEMA'>, 'SELECT': <TokenType.SELECT: 'SELECT'>, 'SEMI': <TokenType.SEMI: 'SEMI'>, 'SET': <TokenType.SET: 'SET'>, 'SETTINGS': <TokenType.SETTINGS: 'SETTINGS'>, 'SHOW': <TokenType.SHOW: 'SHOW'>, 'SIMILAR TO': <TokenType.SIMILAR_TO: 'SIMILAR_TO'>, 'SOME': <TokenType.SOME: 'SOME'>, 'SORT BY': <TokenType.SORT_BY: 'SORT_BY'>, 'START WITH': <TokenType.START_WITH: 'START_WITH'>, 'TABLE': <TokenType.TABLE: 'TABLE'>, 'TABLESAMPLE': <TokenType.TABLE_SAMPLE: 'TABLE_SAMPLE'>, 'TEMP': <TokenType.TEMPORARY: 'TEMPORARY'>, 'TEMPORARY': <TokenType.TEMPORARY: 'TEMPORARY'>, 'THEN': <TokenType.THEN: 'THEN'>, 'TRUE': <TokenType.TRUE: 'TRUE'>, 'UNION': <TokenType.UNION: 'UNION'>, 'UNKNOWN': <TokenType.UNKNOWN: 'UNKNOWN'>, 'UNNEST': <TokenType.UNNEST: 'UNNEST'>, 'UNPIVOT': <TokenType.UNPIVOT: 'UNPIVOT'>, 'UPDATE': <TokenType.UPDATE: 'UPDATE'>, 'USE': <TokenType.USE: 'USE'>, 'USING': <TokenType.USING: 'USING'>, 'UUID': <TokenType.UUID: 'UUID'>, 'VALUES': <TokenType.VALUES: 'VALUES'>, 'VIEW': <TokenType.VIEW: 'VIEW'>, 'VOLATILE': <TokenType.VOLATILE: 'VOLATILE'>, 'WHEN': <TokenType.WHEN: 'WHEN'>, 'WHERE': <TokenType.WHERE: 'WHERE'>, 'WINDOW': <TokenType.WINDOW: 'WINDOW'>, 'WITH': <TokenType.WITH: 'WITH'>, 'APPLY': <TokenType.APPLY: 'APPLY'>, 'ARRAY': <TokenType.ARRAY: 'ARRAY'>, 'BIT': <TokenType.BIT: 'BIT'>, 'BOOL': <TokenType.BOOLEAN: 'BOOLEAN'>, 'BOOLEAN': <TokenType.BOOLEAN: 'BOOLEAN'>, 'BYTE': <TokenType.TINYINT: 'TINYINT'>, 'MEDIUMINT': <TokenType.MEDIUMINT: 'MEDIUMINT'>, 'INT1': <TokenType.TINYINT: 'TINYINT'>, 'TINYINT': <TokenType.TINYINT: 'TINYINT'>, 'INT16': <TokenType.SMALLINT: 'SMALLINT'>, 'SHORT': <TokenType.SMALLINT: 'SMALLINT'>, 'SMALLINT': <TokenType.SMALLINT: 'SMALLINT'>, 'INT128': <TokenType.INT128: 'INT128'>, 'HUGEINT': <TokenType.INT128: 'INT128'>, 'INT2': <TokenType.SMALLINT: 'SMALLINT'>, 'INTEGER': <TokenType.INT: 'INT'>, 'INT': <TokenType.INT: 'INT'>, 'INT4': <TokenType.INT: 'INT'>, 'INT32': <TokenType.INT: 'INT'>, 'INT64': <TokenType.BIGINT: 'BIGINT'>, 'LONG': <TokenType.BIGINT: 'BIGINT'>, 'BIGINT': <TokenType.BIGINT: 'BIGINT'>, 'INT8': <TokenType.TINYINT: 'TINYINT'>, 'DEC': <TokenType.DECIMAL: 'DECIMAL'>, 'DECIMAL': <TokenType.DECIMAL: 'DECIMAL'>, 'BIGDECIMAL': <TokenType.BIGDECIMAL: 'BIGDECIMAL'>, 'BIGNUMERIC': <TokenType.BIGDECIMAL: 'BIGDECIMAL'>, 'MAP': <TokenType.MAP: 'MAP'>, 'NULLABLE': <TokenType.NULLABLE: 'NULLABLE'>, 'NUMBER': <TokenType.DECIMAL: 'DECIMAL'>, 'NUMERIC': <TokenType.DECIMAL: 'DECIMAL'>, 'FIXED': <TokenType.DECIMAL: 'DECIMAL'>, 'REAL': <TokenType.FLOAT: 'FLOAT'>, 'FLOAT': <TokenType.FLOAT: 'FLOAT'>, 'FLOAT4': <TokenType.FLOAT: 'FLOAT'>, 'FLOAT8': <TokenType.DOUBLE: 'DOUBLE'>, 'DOUBLE': <TokenType.DOUBLE: 'DOUBLE'>, 'DOUBLE PRECISION': <TokenType.DOUBLE: 'DOUBLE'>, 'JSON': <TokenType.JSON: 'JSON'>, 'CHAR': <TokenType.TEXT: 'TEXT'>, 'CHARACTER': <TokenType.CHAR: 'CHAR'>, 'NCHAR': <TokenType.NCHAR: 'NCHAR'>, 'VARCHAR': <TokenType.VARCHAR: 'VARCHAR'>, 'VARCHAR2': <TokenType.VARCHAR: 'VARCHAR'>, 'NVARCHAR': <TokenType.NVARCHAR: 'NVARCHAR'>, 'NVARCHAR2': <TokenType.NVARCHAR: 'NVARCHAR'>, 'STR': <TokenType.TEXT: 'TEXT'>, 'STRING': <TokenType.VARCHAR: 'VARCHAR'>, 'TEXT': <TokenType.TEXT: 'TEXT'>, 'LONGTEXT': <TokenType.LONGTEXT: 'LONGTEXT'>, 'MEDIUMTEXT': <TokenType.MEDIUMTEXT: 'MEDIUMTEXT'>, 'TINYTEXT': <TokenType.TINYTEXT: 'TINYTEXT'>, 'CLOB': <TokenType.TEXT: 'TEXT'>, 'LONGVARCHAR': <TokenType.TEXT: 'TEXT'>, 'BINARY': <TokenType.VARBINARY: 'VARBINARY'>, 'BLOB': <TokenType.VARBINARY: 'VARBINARY'>, 'LONGBLOB': <TokenType.LONGBLOB: 'LONGBLOB'>, 'MEDIUMBLOB': <TokenType.MEDIUMBLOB: 'MEDIUMBLOB'>, 'TINYBLOB': <TokenType.TINYBLOB: 'TINYBLOB'>, 'BYTEA': <TokenType.VARBINARY: 'VARBINARY'>, 'VARBINARY': <TokenType.VARBINARY: 'VARBINARY'>, 'TIME': <TokenType.TIME: 'TIME'>, 'TIMETZ': <TokenType.TIMETZ: 'TIMETZ'>, 'TIMESTAMP': <TokenType.TIMESTAMP: 'TIMESTAMP'>, 'TIMESTAMPTZ': <TokenType.TIMESTAMPTZ: 'TIMESTAMPTZ'>, 'TIMESTAMPLTZ': <TokenType.TIMESTAMPLTZ: 'TIMESTAMPLTZ'>, 'DATE': <TokenType.DATE: 'DATE'>, 'DATETIME': <TokenType.DATETIME: 'DATETIME'>, 'INT4RANGE': <TokenType.INT4RANGE: 'INT4RANGE'>, 'INT4MULTIRANGE': <TokenType.INT4MULTIRANGE: 'INT4MULTIRANGE'>, 'INT8RANGE': <TokenType.INT8RANGE: 'INT8RANGE'>, 'INT8MULTIRANGE': <TokenType.INT8MULTIRANGE: 'INT8MULTIRANGE'>, 'NUMRANGE': <TokenType.NUMRANGE: 'NUMRANGE'>, 'NUMMULTIRANGE': <TokenType.NUMMULTIRANGE: 'NUMMULTIRANGE'>, 'TSRANGE': <TokenType.TSRANGE: 'TSRANGE'>, 'TSMULTIRANGE': <TokenType.TSMULTIRANGE: 'TSMULTIRANGE'>, 'TSTZRANGE': <TokenType.TSTZRANGE: 'TSTZRANGE'>, 'TSTZMULTIRANGE': <TokenType.TSTZMULTIRANGE: 'TSTZMULTIRANGE'>, 'DATERANGE': <TokenType.DATERANGE: 'DATERANGE'>, 'DATEMULTIRANGE': <TokenType.DATEMULTIRANGE: 'DATEMULTIRANGE'>, 'UNIQUE': <TokenType.UNIQUE: 'UNIQUE'>, 'STRUCT': <TokenType.STRUCT: 'STRUCT'>, 'VARIANT': <TokenType.VARIANT: 'VARIANT'>, 'ALTER': <TokenType.ALTER: 'ALTER'>, 'ANALYZE': <TokenType.COMMAND: 'COMMAND'>, 'CALL': <TokenType.COMMAND: 'COMMAND'>, 'COMMENT': <TokenType.COMMENT: 'COMMENT'>, 'COPY': <TokenType.COMMAND: 'COMMAND'>, 'EXPLAIN': <TokenType.COMMAND: 'COMMAND'>, 'GRANT': <TokenType.COMMAND: 'COMMAND'>, 'OPTIMIZE': <TokenType.COMMAND: 'COMMAND'>, 'PREPARE': <TokenType.COMMAND: 'COMMAND'>, 'TRUNCATE': <TokenType.COMMAND: 'COMMAND'>, 'VACUUM': <TokenType.COMMAND: 'COMMAND'>, 'USER-DEFINED': <TokenType.USERDEFINED: 'USERDEFINED'>, 'FOR VERSION': <TokenType.VERSION_SNAPSHOT: 'VERSION_SNAPSHOT'>, 'FOR TIMESTAMP': <TokenType.TIMESTAMP_SNAPSHOT: 'TIMESTAMP_SNAPSHOT'>, '//': <TokenType.DIV: 'DIV'>, 'ATTACH': <TokenType.COMMAND: 'COMMAND'>, 'BITSTRING': <TokenType.BIT: 'BIT'>, 'BPCHAR': <TokenType.TEXT: 'TEXT'>, 'CHARACTER VARYING': <TokenType.TEXT: 'TEXT'>, 'EXCLUDE': <TokenType.EXCEPT: 'EXCEPT'>, 'LOGICAL': <TokenType.BOOLEAN: 'BOOLEAN'>, 'PIVOT_WIDER': <TokenType.PIVOT: 'PIVOT'>, 'SIGNED': <TokenType.INT: 'INT'>, 'UBIGINT': <TokenType.UBIGINT: 'UBIGINT'>, 'UINTEGER': <TokenType.UINT: 'UINT'>, 'USMALLINT': <TokenType.USMALLINT: 'USMALLINT'>, 'UTINYINT': <TokenType.UTINYINT: 'UTINYINT'>, 'TIMESTAMP_S': <TokenType.TIMESTAMP_S: 'TIMESTAMP_S'>, 'TIMESTAMP_MS': <TokenType.TIMESTAMP_MS: 'TIMESTAMP_MS'>, 'TIMESTAMP_NS': <TokenType.TIMESTAMP_NS: 'TIMESTAMP_NS'>, 'TIMESTAMP_US': <TokenType.TIMESTAMP: 'TIMESTAMP'>}

##### Inherited Members

sqlglot.tokens.Tokenizer
  Tokenizer
  SINGLE_TOKENS
  BIT_STRINGS
  BYTE_STRINGS
  HEX_STRINGS
  RAW_STRINGS
  HEREDOC_STRINGS
  IDENTIFIERS
  IDENTIFIER_ESCAPES
  QUOTES
  STRING_ESCAPES
  VAR_SINGLE_TOKENS
  WHITE_SPACE
  COMMANDS
  COMMAND_PREFIX_TOKENS
  NUMERIC_LITERALS
  COMMENTS
  dialect
  reset
  tokenize
  peek
  tokenize_rs
  size
  sql
  tokens

class DuckDB.Parser(sqlglot.parser.Parser): View Source

    187    class Parser(parser.Parser):
    188        BITWISE = {
    189            **parser.Parser.BITWISE,
    190            TokenType.TILDA: exp.RegexpLike,
    191        }
    192
    193        FUNCTIONS = {
    194            **parser.Parser.FUNCTIONS,
    195            "ARRAY_HAS": exp.ArrayContains.from_arg_list,
    196            "ARRAY_LENGTH": exp.ArraySize.from_arg_list,
    197            "ARRAY_SORT": exp.SortArray.from_arg_list,
    198            "ARRAY_REVERSE_SORT": _sort_array_reverse,
    199            "DATEDIFF": _parse_date_diff,
    200            "DATE_DIFF": _parse_date_diff,
    201            "DATE_TRUNC": date_trunc_to_time,
    202            "DATETRUNC": date_trunc_to_time,
    203            "DECODE": lambda args: exp.Decode(
    204                this=seq_get(args, 0), charset=exp.Literal.string("utf-8")
    205            ),
    206            "ENCODE": lambda args: exp.Encode(
    207                this=seq_get(args, 0), charset=exp.Literal.string("utf-8")
    208            ),
    209            "EPOCH": exp.TimeToUnix.from_arg_list,
    210            "EPOCH_MS": lambda args: exp.UnixToTime(
    211                this=seq_get(args, 0), scale=exp.UnixToTime.MILLIS
    212            ),
    213            "LIST_HAS": exp.ArrayContains.from_arg_list,
    214            "LIST_REVERSE_SORT": _sort_array_reverse,
    215            "LIST_SORT": exp.SortArray.from_arg_list,
    216            "LIST_VALUE": exp.Array.from_arg_list,
    217            "MAKE_TIMESTAMP": _parse_make_timestamp,
    218            "MEDIAN": lambda args: exp.PercentileCont(
    219                this=seq_get(args, 0), expression=exp.Literal.number(0.5)
    220            ),
    221            "QUANTILE_CONT": exp.PercentileCont.from_arg_list,
    222            "QUANTILE_DISC": exp.PercentileDisc.from_arg_list,
    223            "REGEXP_EXTRACT": lambda args: exp.RegexpExtract(
    224                this=seq_get(args, 0), expression=seq_get(args, 1), group=seq_get(args, 2)
    225            ),
    226            "REGEXP_MATCHES": exp.RegexpLike.from_arg_list,
    227            "REGEXP_REPLACE": lambda args: exp.RegexpReplace(
    228                this=seq_get(args, 0),
    229                expression=seq_get(args, 1),
    230                replacement=seq_get(args, 2),
    231                modifiers=seq_get(args, 3),
    232            ),
    233            "STRFTIME": format_time_lambda(exp.TimeToStr, "duckdb"),
    234            "STRING_SPLIT": exp.Split.from_arg_list,
    235            "STRING_SPLIT_REGEX": exp.RegexpSplit.from_arg_list,
    236            "STRING_TO_ARRAY": exp.Split.from_arg_list,
    237            "STRPTIME": format_time_lambda(exp.StrToTime, "duckdb"),
    238            "STRUCT_PACK": exp.Struct.from_arg_list,
    239            "STR_SPLIT": exp.Split.from_arg_list,
    240            "STR_SPLIT_REGEX": exp.RegexpSplit.from_arg_list,
    241            "TO_TIMESTAMP": exp.UnixToTime.from_arg_list,
    242            "UNNEST": exp.Explode.from_arg_list,
    243            "XOR": binary_from_function(exp.BitwiseXor),
    244        }
    245
    246        FUNCTION_PARSERS = parser.Parser.FUNCTION_PARSERS.copy()
    247        FUNCTION_PARSERS.pop("DECODE", None)
    248
    249        TABLE_ALIAS_TOKENS = parser.Parser.TABLE_ALIAS_TOKENS - {
    250            TokenType.SEMI,
    251            TokenType.ANTI,
    252        }
    253
    254        def _parse_types(
    255            self, check_func: bool = False, schema: bool = False, allow_identifiers: bool = True
    256        ) -> t.Optional[exp.Expression]:
    257            this = super()._parse_types(
    258                check_func=check_func, schema=schema, allow_identifiers=allow_identifiers
    259            )
    260
    261            # DuckDB treats NUMERIC and DECIMAL without precision as DECIMAL(18, 3)
    262            # See: https://duckdb.org/docs/sql/data_types/numeric
    263            if (
    264                isinstance(this, exp.DataType)
    265                and this.is_type("numeric", "decimal")
    266                and not this.expressions
    267            ):
    268                return exp.DataType.build("DECIMAL(18, 3)")
    269
    270            return this
    271
    272        def _parse_struct_types(self) -> t.Optional[exp.Expression]:
    273            return self._parse_field_def()
    274
    275        def _pivot_column_names(self, aggregations: t.List[exp.Expression]) -> t.List[str]:
    276            if len(aggregations) == 1:
    277                return super()._pivot_column_names(aggregations)
    278            return pivot_column_names(aggregations, dialect="duckdb")

Parser consumes a list of tokens produced by the Tokenizer and produces a parsed syntax tree.

###### Arguments:

- error_level: The desired error level. Default: ErrorLevel.IMMEDIATE
- error_message_context: Determines the amount of context to capture from a query string when displaying the error message (in number of characters). Default: 100
- max_errors: Maximum number of error messages to include in a raised ParseError. This is only relevant if error_level is ErrorLevel.RAISE. Default: 3

BITWISE = {<TokenType.AMP: 'AMP'>: <class 'sqlglot.expressions.BitwiseAnd'>, <TokenType.CARET: 'CARET'>: <class 'sqlglot.expressions.BitwiseXor'>, <TokenType.PIPE: 'PIPE'>: <class 'sqlglot.expressions.BitwiseOr'>, <TokenType.TILDA: 'TILDA'>: <class 'sqlglot.expressions.RegexpLike'>}
FUNCTIONS = {'ABS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Abs'>>, 'ANY_VALUE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.AnyValue'>>, 'APPROX_DISTINCT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ApproxDistinct'>>, 'APPROX_COUNT_DISTINCT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ApproxDistinct'>>, 'APPROX_QUANTILE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ApproxQuantile'>>, 'APPROX_TOP_K': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ApproxTopK'>>, 'ARG_MAX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMax'>>, 'ARGMAX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMax'>>, 'MAX_BY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMax'>>, 'ARG_MIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMin'>>, 'ARGMIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMin'>>, 'MIN_BY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMin'>>, 'ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Array'>>, 'ARRAY_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayAgg'>>, 'ARRAY_ALL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayAll'>>, 'ARRAY_ANY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayAny'>>, 'ARRAY_CONCAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayConcat'>>, 'ARRAY_CAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayConcat'>>, 'ARRAY_CONTAINS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayContains'>>, 'FILTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayFilter'>>, 'ARRAY_FILTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayFilter'>>, 'ARRAY_JOIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayJoin'>>, 'ARRAY_SIZE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArraySize'>>, 'ARRAY_SORT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SortArray'>>, 'ARRAY_SUM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArraySum'>>, 'ARRAY_UNION_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayUnionAgg'>>, 'ARRAY_UNIQUE_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayUniqueAgg'>>, 'AVG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Avg'>>, 'CASE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Case'>>, 'CAST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Cast'>>, 'CAST_TO_STR_TYPE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CastToStrType'>>, 'CEIL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Ceil'>>, 'CEILING': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Ceil'>>, 'CHR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Chr'>>, 'CHAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Chr'>>, 'COALESCE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Coalesce'>>, 'IFNULL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Coalesce'>>, 'NVL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Coalesce'>>, 'COLLATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Collate'>>, 'CONCAT': <function Parser.<lambda>>, 'CONCAT_WS': <function Parser.<lambda>>, 'COUNT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Count'>>, 'COUNT_IF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CountIf'>>, 'CURRENT_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentDate'>>, 'CURRENT_DATETIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentDatetime'>>, 'CURRENT_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentTime'>>, 'CURRENT_TIMESTAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentTimestamp'>>, 'CURRENT_USER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentUser'>>, 'DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Date'>>, 'DATE_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateAdd'>>, 'DATEDIFF': <function _parse_date_diff>, 'DATE_DIFF': <function _parse_date_diff>, 'DATEFROMPARTS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateFromParts'>>, 'DATE_STR_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateStrToDate'>>, 'DATE_SUB': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateSub'>>, 'DATE_TO_DATE_STR': <function Parser.<lambda>>, 'DATE_TO_DI': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateToDi'>>, 'DATE_TRUNC': <function date_trunc_to_time>, 'DATETIME_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DatetimeAdd'>>, 'DATETIME_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DatetimeDiff'>>, 'DATETIME_SUB': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DatetimeSub'>>, 'DATETIME_TRUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DatetimeTrunc'>>, 'DAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Day'>>, 'DAY_OF_MONTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfMonth'>>, 'DAYOFMONTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfMonth'>>, 'DAY_OF_WEEK': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfWeek'>>, 'DAYOFWEEK': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfWeek'>>, 'DAY_OF_YEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfYear'>>, 'DAYOFYEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfYear'>>, 'DECODE': <function DuckDB.Parser.<lambda>>, 'DI_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DiToDate'>>, 'ENCODE': <function DuckDB.Parser.<lambda>>, 'EXP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Exp'>>, 'EXPLODE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Explode'>>, 'EXPLODE_OUTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ExplodeOuter'>>, 'EXTRACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Extract'>>, 'FIRST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.First'>>, 'FLATTEN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Flatten'>>, 'FLOOR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Floor'>>, 'FROM_BASE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.FromBase'>>, 'FROM_BASE64': <bound method Func.from_arg_list of <class 'sqlglot.expressions.FromBase64'>>, 'GENERATE_SERIES': <bound method Func.from_arg_list of <class 'sqlglot.expressions.GenerateSeries'>>, 'GREATEST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Greatest'>>, 'GROUP_CONCAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.GroupConcat'>>, 'HEX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Hex'>>, 'HLL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Hll'>>, 'IF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.If'>>, 'INITCAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Initcap'>>, 'IS_INF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.IsInf'>>, 'ISINF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.IsInf'>>, 'IS_NAN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.IsNan'>>, 'ISNAN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.IsNan'>>, 'J_S_O_N_ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONArray'>>, 'J_S_O_N_ARRAY_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONArrayAgg'>>, 'JSON_ARRAY_CONTAINS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONArrayContains'>>, 'JSONB_EXTRACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONBExtract'>>, 'JSONB_EXTRACT_SCALAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONBExtractScalar'>>, 'JSON_EXTRACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONExtract'>>, 'JSON_EXTRACT_SCALAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONExtractScalar'>>, 'JSON_FORMAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONFormat'>>, 'J_S_O_N_OBJECT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONObject'>>, 'J_S_O_N_TABLE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONTable'>>, 'LAST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Last'>>, 'LAST_DATE_OF_MONTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LastDateOfMonth'>>, 'LEAST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Least'>>, 'LEFT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Left'>>, 'LENGTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Length'>>, 'LEN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Length'>>, 'LEVENSHTEIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Levenshtein'>>, 'LN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Ln'>>, 'LOG': <function parse_logarithm>, 'LOG10': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Log10'>>, 'LOG2': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Log2'>>, 'LOGICAL_AND': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalAnd'>>, 'BOOL_AND': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalAnd'>>, 'BOOLAND_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalAnd'>>, 'LOGICAL_OR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalOr'>>, 'BOOL_OR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalOr'>>, 'BOOLOR_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalOr'>>, 'LOWER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Lower'>>, 'LCASE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Lower'>>, 'MD5': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MD5'>>, 'MD5_DIGEST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MD5Digest'>>, 'MAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Map'>>, 'MAP_FROM_ENTRIES': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MapFromEntries'>>, 'MATCH_AGAINST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MatchAgainst'>>, 'MAX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Max'>>, 'MIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Min'>>, 'MONTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Month'>>, 'MONTHS_BETWEEN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MonthsBetween'>>, 'NEXT_VALUE_FOR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.NextValueFor'>>, 'NULLIF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Nullif'>>, 'NUMBER_TO_STR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.NumberToStr'>>, 'NVL2': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Nvl2'>>, 'OPEN_J_S_O_N': <bound method Func.from_arg_list of <class 'sqlglot.expressions.OpenJSON'>>, 'PARAMETERIZED_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ParameterizedAgg'>>, 'PARSE_JSON': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ParseJSON'>>, 'JSON_PARSE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ParseJSON'>>, 'PERCENTILE_CONT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.PercentileCont'>>, 'PERCENTILE_DISC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.PercentileDisc'>>, 'POSEXPLODE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Posexplode'>>, 'POSEXPLODE_OUTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.PosexplodeOuter'>>, 'POWER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Pow'>>, 'POW': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Pow'>>, 'PREDICT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Predict'>>, 'QUANTILE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Quantile'>>, 'RANGE_N': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RangeN'>>, 'READ_CSV': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ReadCSV'>>, 'REDUCE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Reduce'>>, 'REGEXP_EXTRACT': <function DuckDB.Parser.<lambda>>, 'REGEXP_I_LIKE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpILike'>>, 'REGEXP_LIKE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpLike'>>, 'REGEXP_REPLACE': <function DuckDB.Parser.<lambda>>, 'REGEXP_SPLIT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpSplit'>>, 'REPEAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Repeat'>>, 'RIGHT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Right'>>, 'ROUND': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Round'>>, 'ROW_NUMBER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RowNumber'>>, 'SHA': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SHA'>>, 'SHA1': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SHA'>>, 'SHA2': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SHA2'>>, 'SAFE_DIVIDE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SafeDivide'>>, 'SORT_ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SortArray'>>, 'SPLIT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Split'>>, 'SQRT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Sqrt'>>, 'STANDARD_HASH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StandardHash'>>, 'STAR_MAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StarMap'>>, 'STARTS_WITH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StartsWith'>>, 'STARTSWITH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StartsWith'>>, 'STDDEV': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Stddev'>>, 'STDDEV_POP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StddevPop'>>, 'STDDEV_SAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StddevSamp'>>, 'STR_POSITION': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrPosition'>>, 'STR_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrToDate'>>, 'STR_TO_MAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrToMap'>>, 'STR_TO_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrToTime'>>, 'STR_TO_UNIX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrToUnix'>>, 'STRUCT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Struct'>>, 'STRUCT_EXTRACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StructExtract'>>, 'STUFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Stuff'>>, 'INSERT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Stuff'>>, 'SUBSTRING': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Substring'>>, 'SUM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Sum'>>, 'TIME_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeAdd'>>, 'TIME_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeDiff'>>, 'TIME_STR_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeStrToDate'>>, 'TIME_STR_TO_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeStrToTime'>>, 'TIME_STR_TO_UNIX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeStrToUnix'>>, 'TIME_SUB': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeSub'>>, 'TIME_TO_STR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeToStr'>>, 'TIME_TO_TIME_STR': <function Parser.<lambda>>, 'TIME_TO_UNIX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeToUnix'>>, 'TIME_TRUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeTrunc'>>, 'TIMESTAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Timestamp'>>, 'TIMESTAMP_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampAdd'>>, 'TIMESTAMP_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampDiff'>>, 'TIMESTAMP_FROM_PARTS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampFromParts'>>, 'TIMESTAMP_SUB': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampSub'>>, 'TIMESTAMP_TRUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampTrunc'>>, 'TO_BASE64': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToBase64'>>, 'TO_CHAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToChar'>>, 'TO_DAYS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToDays'>>, 'TRANSFORM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Transform'>>, 'TRIM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Trim'>>, 'TRY_CAST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TryCast'>>, 'TS_OR_DI_TO_DI': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDiToDi'>>, 'TS_OR_DS_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDsAdd'>>, 'TS_OR_DS_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDsDiff'>>, 'TS_OR_DS_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDsToDate'>>, 'TS_OR_DS_TO_DATE_STR': <function Parser.<lambda>>, 'UNHEX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Unhex'>>, 'UNIX_TO_STR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.UnixToStr'>>, 'UNIX_TO_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.UnixToTime'>>, 'UNIX_TO_TIME_STR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.UnixToTimeStr'>>, 'UPPER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Upper'>>, 'UCASE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Upper'>>, 'VAR_MAP': <function parse_var_map>, 'VARIANCE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Variance'>>, 'VARIANCE_SAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Variance'>>, 'VAR_SAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Variance'>>, 'VARIANCE_POP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.VariancePop'>>, 'VAR_POP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.VariancePop'>>, 'WEEK': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Week'>>, 'WEEK_OF_YEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.WeekOfYear'>>, 'WEEKOFYEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.WeekOfYear'>>, 'WHEN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.When'>>, 'X_M_L_TABLE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.XMLTable'>>, 'XOR': <function binary_from_function.<locals>.<lambda>>, 'YEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Year'>>, 'GLOB': <function Parser.<lambda>>, 'LIKE': <function parse_like>, 'ARRAY_HAS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayContains'>>, 'ARRAY_LENGTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArraySize'>>, 'ARRAY_REVERSE_SORT': <function _sort_array_reverse>, 'DATETRUNC': <function date_trunc_to_time>, 'EPOCH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeToUnix'>>, 'EPOCH_MS': <function DuckDB.Parser.<lambda>>, 'LIST_HAS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayContains'>>, 'LIST_REVERSE_SORT': <function _sort_array_reverse>, 'LIST_SORT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SortArray'>>, 'LIST_VALUE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Array'>>, 'MAKE_TIMESTAMP': <function _parse_make_timestamp>, 'MEDIAN': <function DuckDB.Parser.<lambda>>, 'QUANTILE_CONT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.PercentileCont'>>, 'QUANTILE_DISC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.PercentileDisc'>>, 'REGEXP_MATCHES': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpLike'>>, 'STRFTIME': <function format_time_lambda.<locals>._format_time>, 'STRING_SPLIT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Split'>>, 'STRING_SPLIT_REGEX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpSplit'>>, 'STRING_TO_ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Split'>>, 'STRPTIME': <function format_time_lambda.<locals>._format_time>, 'STRUCT_PACK': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Struct'>>, 'STR_SPLIT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Split'>>, 'STR_SPLIT_REGEX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpSplit'>>, 'TO_TIMESTAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.UnixToTime'>>, 'UNNEST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Explode'>>}
FUNCTION_PARSERS = {'ANY_VALUE': <function Parser.<lambda>>, 'CAST': <function Parser.<lambda>>, 'CONVERT': <function Parser.<lambda>>, 'EXTRACT': <function Parser.<lambda>>, 'JSON_OBJECT': <function Parser.<lambda>>, 'JSON_TABLE': <function Parser.<lambda>>, 'MATCH': <function Parser.<lambda>>, 'OPENJSON': <function Parser.<lambda>>, 'POSITION': <function Parser.<lambda>>, 'PREDICT': <function Parser.<lambda>>, 'SAFE_CAST': <function Parser.<lambda>>, 'STRING_AGG': <function Parser.<lambda>>, 'SUBSTRING': <function Parser.<lambda>>, 'TRIM': <function Parser.<lambda>>, 'TRY_CAST': <function Parser.<lambda>>, 'TRY_CONVERT': <function Parser.<lambda>>}
TABLE_ALIAS_TOKENS = {<TokenType.TIMESTAMP_MS: 'TIMESTAMP_MS'>, <TokenType.INT4MULTIRANGE: 'INT4MULTIRANGE'>, <TokenType.AUTO_INCREMENT: 'AUTO_INCREMENT'>, <TokenType.USERDEFINED: 'USERDEFINED'>, <TokenType.VAR: 'VAR'>, <TokenType.TINYINT: 'TINYINT'>, <TokenType.DATE: 'DATE'>, <TokenType.CASE: 'CASE'>, <TokenType.MONEY: 'MONEY'>, <TokenType.HSTORE: 'HSTORE'>, <TokenType.SMALLMONEY: 'SMALLMONEY'>, <TokenType.OVERLAPS: 'OVERLAPS'>, <TokenType.TEXT: 'TEXT'>, <TokenType.COLLATE: 'COLLATE'>, <TokenType.REFERENCES: 'REFERENCES'>, <TokenType.UPDATE: 'UPDATE'>, <TokenType.UUID: 'UUID'>, <TokenType.INT4RANGE: 'INT4RANGE'>, <TokenType.TINYBLOB: 'TINYBLOB'>, <TokenType.BIGDECIMAL: 'BIGDECIMAL'>, <TokenType.FIRST: 'FIRST'>, <TokenType.TIMESTAMP_S: 'TIMESTAMP_S'>, <TokenType.ISNULL: 'ISNULL'>, <TokenType.DOUBLE: 'DOUBLE'>, <TokenType.DEFAULT: 'DEFAULT'>, <TokenType.ROWS: 'ROWS'>, <TokenType.NUMRANGE: 'NUMRANGE'>, <TokenType.INDEX: 'INDEX'>, <TokenType.DATABASE: 'DATABASE'>, <TokenType.INT128: 'INT128'>, <TokenType.ALL: 'ALL'>, <TokenType.SET: 'SET'>, <TokenType.JSONB: 'JSONB'>, <TokenType.CACHE: 'CACHE'>, <TokenType.NULL: 'NULL'>, <TokenType.BOOLEAN: 'BOOLEAN'>, <TokenType.OPERATOR: 'OPERATOR'>, <TokenType.TSTZMULTIRANGE: 'TSTZMULTIRANGE'>, <TokenType.VARBINARY: 'VARBINARY'>, <TokenType.COMMENT: 'COMMENT'>, <TokenType.ROW: 'ROW'>, <TokenType.BIT: 'BIT'>, <TokenType.ENUM: 'ENUM'>, <TokenType.UDECIMAL: 'UDECIMAL'>, <TokenType.FALSE: 'FALSE'>, <TokenType.TIMESTAMP: 'TIMESTAMP'>, <TokenType.IMAGE: 'IMAGE'>, <TokenType.HLLSKETCH: 'HLLSKETCH'>, <TokenType.REFRESH: 'REFRESH'>, <TokenType.BIGINT: 'BIGINT'>, <TokenType.DATERANGE: 'DATERANGE'>, <TokenType.OBJECT_IDENTIFIER: 'OBJECT_IDENTIFIER'>, <TokenType.NEXT: 'NEXT'>, <TokenType.SOME: 'SOME'>, <TokenType.FUNCTION: 'FUNCTION'>, <TokenType.TEMPORARY: 'TEMPORARY'>, <TokenType.ASC: 'ASC'>, <TokenType.IPADDRESS: 'IPADDRESS'>, <TokenType.ARRAY: 'ARRAY'>, <TokenType.TINYTEXT: 'TINYTEXT'>, <TokenType.TIMESTAMPTZ: 'TIMESTAMPTZ'>, <TokenType.XML: 'XML'>, <TokenType.LOAD: 'LOAD'>, <TokenType.TIMESTAMP_NS: 'TIMESTAMP_NS'>, <TokenType.TIMETZ: 'TIMETZ'>, <TokenType.DATEMULTIRANGE: 'DATEMULTIRANGE'>, <TokenType.UNIQUEIDENTIFIER: 'UNIQUEIDENTIFIER'>, <TokenType.INTERVAL: 'INTERVAL'>, <TokenType.FIXEDSTRING: 'FIXEDSTRING'>, <TokenType.VARIANT: 'VARIANT'>, <TokenType.BINARY: 'BINARY'>, <TokenType.UMEDIUMINT: 'UMEDIUMINT'>, <TokenType.MERGE: 'MERGE'>, <TokenType.LONGBLOB: 'LONGBLOB'>, <TokenType.STRUCT: 'STRUCT'>, <TokenType.JSON: 'JSON'>, <TokenType.COLUMN: 'COLUMN'>, <TokenType.PSEUDO_TYPE: 'PSEUDO_TYPE'>, <TokenType.NESTED: 'NESTED'>, <TokenType.OBJECT: 'OBJECT'>, <TokenType.SETTINGS: 'SETTINGS'>, <TokenType.SMALLINT: 'SMALLINT'>, <TokenType.UNIQUE: 'UNIQUE'>, <TokenType.CONSTRAINT: 'CONSTRAINT'>, <TokenType.NULLABLE: 'NULLABLE'>, <TokenType.BEGIN: 'BEGIN'>, <TokenType.UINT: 'UINT'>, <TokenType.INT8MULTIRANGE: 'INT8MULTIRANGE'>, <TokenType.SHOW: 'SHOW'>, <TokenType.OVERWRITE: 'OVERWRITE'>, <TokenType.UNPIVOT: 'UNPIVOT'>, <TokenType.USE: 'USE'>, <TokenType.GEOMETRY: 'GEOMETRY'>, <TokenType.VOLATILE: 'VOLATILE'>, <TokenType.MODEL: 'MODEL'>, <TokenType.PROCEDURE: 'PROCEDURE'>, <TokenType.UINT256: 'UINT256'>, <TokenType.MEDIUMTEXT: 'MEDIUMTEXT'>, <TokenType.INT8RANGE: 'INT8RANGE'>, <TokenType.RANGE: 'RANGE'>, <TokenType.ESCAPE: 'ESCAPE'>, <TokenType.DESC: 'DESC'>, <TokenType.ANY: 'ANY'>, <TokenType.INET: 'INET'>, <TokenType.ENUM8: 'ENUM8'>, <TokenType.DECIMAL: 'DECIMAL'>, <TokenType.COMMIT: 'COMMIT'>, <TokenType.CURRENT_TIME: 'CURRENT_TIME'>, <TokenType.MAP: 'MAP'>, <TokenType.KILL: 'KILL'>, <TokenType.CURRENT_USER: 'CURRENT_USER'>, <TokenType.TIME: 'TIME'>, <TokenType.CURRENT_DATETIME: 'CURRENT_DATETIME'>, <TokenType.TIMESTAMPLTZ: 'TIMESTAMPLTZ'>, <TokenType.UBIGINT: 'UBIGINT'>, <TokenType.NVARCHAR: 'NVARCHAR'>, <TokenType.FOREIGN_KEY: 'FOREIGN_KEY'>, <TokenType.FLOAT: 'FLOAT'>, <TokenType.SERIAL: 'SERIAL'>, <TokenType.DESCRIBE: 'DESCRIBE'>, <TokenType.MEDIUMINT: 'MEDIUMINT'>, <TokenType.DIV: 'DIV'>, <TokenType.IS: 'IS'>, <TokenType.KEEP: 'KEEP'>, <TokenType.NCHAR: 'NCHAR'>, <TokenType.FILTER: 'FILTER'>, <TokenType.ENUM16: 'ENUM16'>, <TokenType.RECURSIVE: 'RECURSIVE'>, <TokenType.YEAR: 'YEAR'>, <TokenType.CHAR: 'CHAR'>, <TokenType.UTINYINT: 'UTINYINT'>, <TokenType.LOWCARDINALITY: 'LOWCARDINALITY'>, <TokenType.INT256: 'INT256'>, <TokenType.DATETIME: 'DATETIME'>, <TokenType.IPPREFIX: 'IPPREFIX'>, <TokenType.MEDIUMBLOB: 'MEDIUMBLOB'>, <TokenType.NUMMULTIRANGE: 'NUMMULTIRANGE'>, <TokenType.PIVOT: 'PIVOT'>, <TokenType.UNKNOWN: 'UNKNOWN'>, <TokenType.ORDINALITY: 'ORDINALITY'>, <TokenType.TRUE: 'TRUE'>, <TokenType.LONGTEXT: 'LONGTEXT'>, <TokenType.BIGSERIAL: 'BIGSERIAL'>, <TokenType.PERCENT: 'PERCENT'>, <TokenType.CURRENT_TIMESTAMP: 'CURRENT_TIMESTAMP'>, <TokenType.END: 'END'>, <TokenType.TABLE: 'TABLE'>, <TokenType.UINT128: 'UINT128'>, <TokenType.TOP: 'TOP'>, <TokenType.CURRENT_DATE: 'CURRENT_DATE'>, <TokenType.DELETE: 'DELETE'>, <TokenType.REPLACE: 'REPLACE'>, <TokenType.GEOGRAPHY: 'GEOGRAPHY'>, <TokenType.PARTITION: 'PARTITION'>, <TokenType.INT: 'INT'>, <TokenType.EXECUTE: 'EXECUTE'>, <TokenType.COMMAND: 'COMMAND'>, <TokenType.DICTIONARY: 'DICTIONARY'>, <TokenType.FORMAT: 'FORMAT'>, <TokenType.SUPER: 'SUPER'>, <TokenType.TSTZRANGE: 'TSTZRANGE'>, <TokenType.DATETIME64: 'DATETIME64'>, <TokenType.VIEW: 'VIEW'>, <TokenType.TSMULTIRANGE: 'TSMULTIRANGE'>, <TokenType.SMALLSERIAL: 'SMALLSERIAL'>, <TokenType.ROWVERSION: 'ROWVERSION'>, <TokenType.TSRANGE: 'TSRANGE'>, <TokenType.SCHEMA: 'SCHEMA'>, <TokenType.VARCHAR: 'VARCHAR'>, <TokenType.PRAGMA: 'PRAGMA'>, <TokenType.EXISTS: 'EXISTS'>, <TokenType.USMALLINT: 'USMALLINT'>}
SHOW_TRIE: Dict = {}
SET_TRIE: Dict = {'GLOBAL': {0: True}, 'LOCAL': {0: True}, 'SESSION': {0: True}, 'TRANSACTION': {0: True}}

##### Inherited Members

sqlglot.parser.Parser
  Parser
  NO_PAREN_FUNCTIONS
  STRUCT_TYPE_TOKENS
  NESTED_TYPE_TOKENS
  ENUM_TYPE_TOKENS
  TYPE_TOKENS
  SIGNED_TO_UNSIGNED_TYPE_TOKEN
  SUBQUERY_PREDICATES
  RESERVED_TOKENS
  DB_CREATABLES
  CREATABLES
  ID_VAR_TOKENS
  INTERVAL_VARS
  COMMENT_TABLE_ALIAS_TOKENS
  UPDATE_ALIAS_TOKENS
  TRIM_TYPES
  FUNC_TOKENS
  CONJUNCTION
  EQUALITY
  COMPARISON
  TERM
  FACTOR
  EXPONENT
  TIMES
  TIMESTAMPS
  SET_OPERATIONS
  JOIN_METHODS
  JOIN_SIDES
  JOIN_KINDS
  JOIN_HINTS
  LAMBDAS
  COLUMN_OPERATORS
  EXPRESSION_PARSERS
  STATEMENT_PARSERS
  UNARY_PARSERS
  PRIMARY_PARSERS
  PLACEHOLDER_PARSERS
  RANGE_PARSERS
  PROPERTY_PARSERS
  CONSTRAINT_PARSERS
  ALTER_PARSERS
  SCHEMA_UNNAMED_CONSTRAINTS
  NO_PAREN_FUNCTION_PARSERS
  INVALID_FUNC_NAME_TOKENS
  FUNCTIONS_WITH_ALIASED_ARGS
  QUERY_MODIFIER_PARSERS
  SET_PARSERS
  SHOW_PARSERS
  TYPE_LITERAL_PARSERS
  MODIFIABLES
  DDL_SELECT_TOKENS
  PRE_VOLATILE_TOKENS
  TRANSACTION_KIND
  TRANSACTION_CHARACTERISTICS
  INSERT_ALTERNATIVES
  CLONE_KEYWORDS
  HISTORICAL_DATA_KIND
  OPCLASS_FOLLOW_KEYWORDS
  OPTYPE_FOLLOW_TOKENS
  TABLE_INDEX_HINT_TOKENS
  WINDOW_ALIAS_TOKENS
  WINDOW_BEFORE_PAREN_TOKENS
  WINDOW_SIDES
  FETCH_TOKENS
  ADD_CONSTRAINT_TOKENS
  DISTINCT_TOKENS
  NULL_TOKENS
  UNNEST_OFFSET_ALIAS_TOKENS
  STRICT_CAST
  PREFIXED_PIVOT_COLUMNS
  IDENTIFY_PIVOT_STRINGS
  LOG_DEFAULTS_TO_LN
  ALTER_TABLE_ADD_REQUIRED_FOR_EACH_COLUMN
  TABLESAMPLE_CSV
  SET_REQUIRES_ASSIGNMENT_DELIMITER
  TRIM_PATTERN_FIRST
  MODIFIERS_ATTACHED_TO_UNION
  UNION_MODIFIERS
  error_level
  error_message_context
  max_errors
  dialect
  reset
  parse
  parse_into
  check_errors
  raise_error
  expression
  validate_expression
  errors
  sql

class DuckDB.Generator(sqlglot.generator.Generator): View Source

    280    class Generator(generator.Generator):
    281        JOIN_HINTS = False
    282        TABLE_HINTS = False
    283        QUERY_HINTS = False
    284        LIMIT_FETCH = "LIMIT"
    285        STRUCT_DELIMITER = ("(", ")")
    286        RENAME_TABLE_WITH_DB = False
    287        NVL2_SUPPORTED = False
    288        SEMI_ANTI_JOIN_WITH_SIDE = False
    289
    290        TRANSFORMS = {
    291            **generator.Generator.TRANSFORMS,
    292            exp.ApproxDistinct: approx_count_distinct_sql,
    293            exp.Array: lambda self, e: self.func("ARRAY", e.expressions[0])
    294            if e.expressions and e.expressions[0].find(exp.Select)
    295            else inline_array_sql(self, e),
    296            exp.ArraySize: rename_func("ARRAY_LENGTH"),
    297            exp.ArgMax: arg_max_or_min_no_count("ARG_MAX"),
    298            exp.ArgMin: arg_max_or_min_no_count("ARG_MIN"),
    299            exp.ArraySort: _array_sort_sql,
    300            exp.ArraySum: rename_func("LIST_SUM"),
    301            exp.BitwiseXor: rename_func("XOR"),
    302            exp.CommentColumnConstraint: no_comment_column_constraint_sql,
    303            exp.CurrentDate: lambda self, e: "CURRENT_DATE",
    304            exp.CurrentTime: lambda self, e: "CURRENT_TIME",
    305            exp.CurrentTimestamp: lambda self, e: "CURRENT_TIMESTAMP",
    306            exp.DayOfMonth: rename_func("DAYOFMONTH"),
    307            exp.DayOfWeek: rename_func("DAYOFWEEK"),
    308            exp.DayOfYear: rename_func("DAYOFYEAR"),
    309            exp.DataType: _datatype_sql,
    310            exp.Date: _date_sql,
    311            exp.DateAdd: _date_delta_sql,
    312            exp.DateFromParts: rename_func("MAKE_DATE"),
    313            exp.DateSub: _date_delta_sql,
    314            exp.DateDiff: lambda self, e: self.func(
    315                "DATE_DIFF", f"'{e.args.get('unit') or 'day'}'", e.expression, e.this
    316            ),
    317            exp.DateStrToDate: datestrtodate_sql,
    318            exp.DateToDi: lambda self, e: f"CAST(STRFTIME({self.sql(e, 'this')}, {DuckDB.DATEINT_FORMAT}) AS INT)",
    319            exp.Decode: lambda self, e: encode_decode_sql(self, e, "DECODE", replace=False),
    320            exp.DiToDate: lambda self, e: f"CAST(STRPTIME(CAST({self.sql(e, 'this')} AS TEXT), {DuckDB.DATEINT_FORMAT}) AS DATE)",
    321            exp.Encode: lambda self, e: encode_decode_sql(self, e, "ENCODE", replace=False),
    322            exp.Explode: rename_func("UNNEST"),
    323            exp.IntDiv: lambda self, e: self.binary(e, "//"),
    324            exp.IsInf: rename_func("ISINF"),
    325            exp.IsNan: rename_func("ISNAN"),
    326            exp.JSONExtract: arrow_json_extract_sql,
    327            exp.JSONExtractScalar: arrow_json_extract_scalar_sql,
    328            exp.JSONFormat: _json_format_sql,
    329            exp.JSONBExtract: arrow_json_extract_sql,
    330            exp.JSONBExtractScalar: arrow_json_extract_scalar_sql,
    331            exp.LogicalOr: rename_func("BOOL_OR"),
    332            exp.LogicalAnd: rename_func("BOOL_AND"),
    333            exp.MonthsBetween: lambda self, e: self.func(
    334                "DATEDIFF",
    335                "'month'",
    336                exp.cast(e.expression, "timestamp", copy=True),
    337                exp.cast(e.this, "timestamp", copy=True),
    338            ),
    339            exp.ParseJSON: rename_func("JSON"),
    340            exp.PercentileCont: rename_func("QUANTILE_CONT"),
    341            exp.PercentileDisc: rename_func("QUANTILE_DISC"),
    342            # DuckDB doesn't allow qualified columns inside of PIVOT expressions.
    343            # See: https://github.com/duckdb/duckdb/blob/671faf92411182f81dce42ac43de8bfb05d9909e/src/planner/binder/tableref/bind_pivot.cpp#L61-L62
    344            exp.Pivot: transforms.preprocess([transforms.unqualify_columns]),
    345            exp.Properties: no_properties_sql,
    346            exp.RegexpExtract: regexp_extract_sql,
    347            exp.RegexpReplace: lambda self, e: self.func(
    348                "REGEXP_REPLACE",
    349                e.this,
    350                e.expression,
    351                e.args.get("replacement"),
    352                e.args.get("modifiers"),
    353            ),
    354            exp.RegexpLike: rename_func("REGEXP_MATCHES"),
    355            exp.RegexpSplit: rename_func("STR_SPLIT_REGEX"),
    356            exp.SafeDivide: no_safe_divide_sql,
    357            exp.Split: rename_func("STR_SPLIT"),
    358            exp.SortArray: _sort_array_sql,
    359            exp.StrPosition: str_position_sql,
    360            exp.StrToDate: lambda self, e: f"CAST({str_to_time_sql(self, e)} AS DATE)",
    361            exp.StrToTime: str_to_time_sql,
    362            exp.StrToUnix: lambda self, e: f"EPOCH(STRPTIME({self.sql(e, 'this')}, {self.format_time(e)}))",
    363            exp.Struct: _struct_sql,
    364            exp.Timestamp: no_timestamp_sql,
    365            exp.TimestampFromParts: rename_func("MAKE_TIMESTAMP"),
    366            exp.TimestampTrunc: timestamptrunc_sql,
    367            exp.TimeStrToDate: lambda self, e: f"CAST({self.sql(e, 'this')} AS DATE)",
    368            exp.TimeStrToTime: timestrtotime_sql,
    369            exp.TimeStrToUnix: lambda self, e: f"EPOCH(CAST({self.sql(e, 'this')} AS TIMESTAMP))",
    370            exp.TimeToStr: lambda self, e: f"STRFTIME({self.sql(e, 'this')}, {self.format_time(e)})",
    371            exp.TimeToUnix: rename_func("EPOCH"),
    372            exp.TsOrDiToDi: lambda self, e: f"CAST(SUBSTR(REPLACE(CAST({self.sql(e, 'this')} AS TEXT), '-', ''), 1, 8) AS INT)",
    373            exp.TsOrDsAdd: _ts_or_ds_add_sql,
    374            exp.TsOrDsDiff: lambda self, e: self.func(
    375                "DATE_DIFF",
    376                f"'{e.args.get('unit') or 'day'}'",
    377                exp.cast(e.expression, "TIMESTAMP"),
    378                exp.cast(e.this, "TIMESTAMP"),
    379            ),
    380            exp.TsOrDsToDate: ts_or_ds_to_date_sql("duckdb"),
    381            exp.UnixToStr: lambda self, e: f"STRFTIME(TO_TIMESTAMP({self.sql(e, 'this')}), {self.format_time(e)})",
    382            exp.UnixToTime: _unix_to_time_sql,
    383            exp.UnixToTimeStr: lambda self, e: f"CAST(TO_TIMESTAMP({self.sql(e, 'this')}) AS TEXT)",
    384            exp.VariancePop: rename_func("VAR_POP"),
    385            exp.WeekOfYear: rename_func("WEEKOFYEAR"),
    386            exp.Xor: bool_xor_sql,
    387        }
    388
    389        TYPE_MAPPING = {
    390            **generator.Generator.TYPE_MAPPING,
    391            exp.DataType.Type.BINARY: "BLOB",
    392            exp.DataType.Type.CHAR: "TEXT",
    393            exp.DataType.Type.FLOAT: "REAL",
    394            exp.DataType.Type.NCHAR: "TEXT",
    395            exp.DataType.Type.NVARCHAR: "TEXT",
    396            exp.DataType.Type.UINT: "UINTEGER",
    397            exp.DataType.Type.VARBINARY: "BLOB",
    398            exp.DataType.Type.VARCHAR: "TEXT",
    399            exp.DataType.Type.TIMESTAMP_S: "TIMESTAMP_S",
    400            exp.DataType.Type.TIMESTAMP_MS: "TIMESTAMP_MS",
    401            exp.DataType.Type.TIMESTAMP_NS: "TIMESTAMP_NS",
    402        }
    403
    404        STAR_MAPPING = {**generator.Generator.STAR_MAPPING, "except": "EXCLUDE"}
    405
    406        UNWRAPPED_INTERVAL_VALUES = (exp.Column, exp.Literal, exp.Paren)
    407
    408        PROPERTIES_LOCATION = {
    409            **generator.Generator.PROPERTIES_LOCATION,
    410            exp.VolatileProperty: exp.Properties.Location.UNSUPPORTED,
    411        }
    412
    413        def interval_sql(self, expression: exp.Interval) -> str:
    414            multiplier: t.Optional[int] = None
    415            unit = expression.text("unit").lower()
    416
    417            if unit.startswith("week"):
    418                multiplier = 7
    419            if unit.startswith("quarter"):
    420                multiplier = 90
    421
    422            if multiplier:
    423                return f"({multiplier} * {super().interval_sql(exp.Interval(this=expression.this, unit=exp.var('day')))})"
    424
    425            return super().interval_sql(expression)
    426
    427        def tablesample_sql(
    428            self, expression: exp.TableSample, seed_prefix: str = "SEED", sep: str = " AS "
    429        ) -> str:
    430            return super().tablesample_sql(expression, seed_prefix="REPEATABLE", sep=sep)

Generator converts a given syntax tree to the corresponding SQL string.

###### Arguments:

- pretty: Whether or not to format the produced SQL string. Default: False.
- identify: Determines when an identifier should be quoted. Possible values are: False (default): Never quote, except in cases where it's mandatory by the dialect. True or 'always': Always quote. 'safe': Only quote identifiers that are case insensitive.
- normalize: Whether or not to normalize identifiers to lowercase. Default: False.
- pad: Determines the pad size in a formatted string. Default: 2.
- indent: Determines the indentation size in a formatted string. Default: 2.
- normalize_functions: Whether or not to normalize all function names. Possible values are: "upper" or True (default): Convert names to uppercase. "lower": Convert names to lowercase. False: Disables function name normalization.
- unsupported_level: Determines the generator's behavior when it encounters unsupported expressions. Default ErrorLevel.WARN.
- max_unsupported: Maximum number of unsupported messages to include in a raised UnsupportedError. This is only relevant if unsupported_level is ErrorLevel.RAISE. Default: 3
- leading_comma: Determines whether or not the comma is leading or trailing in select expressions. This is only relevant when generating in pretty mode. Default: False
- max_text_width: The max number of characters in a segment before creating new lines in pretty mode. The default is on the smaller end because the length only represents a segment and not the true line length. Default: 80
- comments: Whether or not to preserve comments in the output SQL code. Default: True

JOIN_HINTS = False
TABLE_HINTS = False
QUERY_HINTS = False
LIMIT_FETCH = 'LIMIT'
STRUCT_DELIMITER = ('(', ')')
RENAME_TABLE_WITH_DB = False
NVL2_SUPPORTED = False
SEMI_ANTI_JOIN_WITH_SIDE = False
TRANSFORMS = {<class 'sqlglot.expressions.DateAdd'>: <function _date_delta_sql>, <class 'sqlglot.expressions.CaseSpecificColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CharacterSetColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CharacterSetProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CheckColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ClusteredColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CollateColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CopyGrantsProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CommentColumnConstraint'>: <function no_comment_column_constraint_sql>, <class 'sqlglot.expressions.DateFormatColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.DefaultColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.EncodeColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ExecuteAsProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ExternalProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.HeapProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.InlineLengthColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.InputModelProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.IntervalSpan'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.LanguageProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.LocationProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.LogProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.MaterializedProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.NoPrimaryIndexProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.NonClusteredColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.NotForReplicationColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.OnCommitProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.OnProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.OnUpdateColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.OutputModelProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.PathColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.RemoteWithConnectionModelProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ReturnsProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SampleProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SetProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SettingsProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SqlSecurityProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.StabilityProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.TemporaryProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ToTableProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.TransientProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.TransformModelProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.TitleColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.UppercaseColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.VarMap'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.VolatileProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.WithJournalTableProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ApproxDistinct'>: <function approx_count_distinct_sql>, <class 'sqlglot.expressions.Array'>: <function DuckDB.Generator.<lambda>>, <class 'sqlglot.expressions.ArraySize'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.ArgMax'>: <function arg_max_or_min_no_count.<locals>._arg_max_or_min_sql>, <class 'sqlglot.expressions.ArgMin'>: <function arg_max_or_min_no_count.<locals>._arg_max_or_min_sql>, <class 'sqlglot.expressions.ArraySort'>: <function _array_sort_sql>, <class 'sqlglot.expressions.ArraySum'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.BitwiseXor'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.CurrentDate'>: <function DuckDB.Generator.<lambda>>, <class 'sqlglot.expressions.CurrentTime'>: <function DuckDB.Generator.<lambda>>, <class 'sqlglot.expressions.CurrentTimestamp'>: <function DuckDB.Generator.<lambda>>, <class 'sqlglot.expressions.DayOfMonth'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.DayOfWeek'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.DayOfYear'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.DataType'>: <function _datatype_sql>, <class 'sqlglot.expressions.Date'>: <function _date_sql>, <class 'sqlglot.expressions.DateFromParts'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.DateSub'>: <function _date_delta_sql>, <class 'sqlglot.expressions.DateDiff'>: <function DuckDB.Generator.<lambda>>, <class 'sqlglot.expressions.DateStrToDate'>: <function datestrtodate_sql>, <class 'sqlglot.expressions.DateToDi'>: <function DuckDB.Generator.<lambda>>, <class 'sqlglot.expressions.Decode'>: <function DuckDB.Generator.<lambda>>, <class 'sqlglot.expressions.DiToDate'>: <function DuckDB.Generator.<lambda>>, <class 'sqlglot.expressions.Encode'>: <function DuckDB.Generator.<lambda>>, <class 'sqlglot.expressions.Explode'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.IntDiv'>: <function DuckDB.Generator.<lambda>>, <class 'sqlglot.expressions.IsInf'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.IsNan'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.JSONExtract'>: <function arrow_json_extract_sql>, <class 'sqlglot.expressions.JSONExtractScalar'>: <function arrow_json_extract_scalar_sql>, <class 'sqlglot.expressions.JSONFormat'>: <function _json_format_sql>, <class 'sqlglot.expressions.JSONBExtract'>: <function arrow_json_extract_sql>, <class 'sqlglot.expressions.JSONBExtractScalar'>: <function arrow_json_extract_scalar_sql>, <class 'sqlglot.expressions.LogicalOr'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.LogicalAnd'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.MonthsBetween'>: <function DuckDB.Generator.<lambda>>, <class 'sqlglot.expressions.ParseJSON'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.PercentileCont'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.PercentileDisc'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.Pivot'>: <function preprocess.<locals>._to_sql>, <class 'sqlglot.expressions.Properties'>: <function no_properties_sql>, <class 'sqlglot.expressions.RegexpExtract'>: <function regexp_extract_sql>, <class 'sqlglot.expressions.RegexpReplace'>: <function DuckDB.Generator.<lambda>>, <class 'sqlglot.expressions.RegexpLike'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.RegexpSplit'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.SafeDivide'>: <function no_safe_divide_sql>, <class 'sqlglot.expressions.Split'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.SortArray'>: <function _sort_array_sql>, <class 'sqlglot.expressions.StrPosition'>: <function str_position_sql>, <class 'sqlglot.expressions.StrToDate'>: <function DuckDB.Generator.<lambda>>, <class 'sqlglot.expressions.StrToTime'>: <function str_to_time_sql>, <class 'sqlglot.expressions.StrToUnix'>: <function DuckDB.Generator.<lambda>>, <class 'sqlglot.expressions.Struct'>: <function _struct_sql>, <class 'sqlglot.expressions.Timestamp'>: <function no_timestamp_sql>, <class 'sqlglot.expressions.TimestampFromParts'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.TimestampTrunc'>: <function timestamptrunc_sql>, <class 'sqlglot.expressions.TimeStrToDate'>: <function DuckDB.Generator.<lambda>>, <class 'sqlglot.expressions.TimeStrToTime'>: <function timestrtotime_sql>, <class 'sqlglot.expressions.TimeStrToUnix'>: <function DuckDB.Generator.<lambda>>, <class 'sqlglot.expressions.TimeToStr'>: <function DuckDB.Generator.<lambda>>, <class 'sqlglot.expressions.TimeToUnix'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.TsOrDiToDi'>: <function DuckDB.Generator.<lambda>>, <class 'sqlglot.expressions.TsOrDsAdd'>: <function _ts_or_ds_add_sql>, <class 'sqlglot.expressions.TsOrDsDiff'>: <function DuckDB.Generator.<lambda>>, <class 'sqlglot.expressions.TsOrDsToDate'>: <function ts_or_ds_to_date_sql.<locals>._ts_or_ds_to_date_sql>, <class 'sqlglot.expressions.UnixToStr'>: <function DuckDB.Generator.<lambda>>, <class 'sqlglot.expressions.UnixToTime'>: <function _unix_to_time_sql>, <class 'sqlglot.expressions.UnixToTimeStr'>: <function DuckDB.Generator.<lambda>>, <class 'sqlglot.expressions.VariancePop'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.WeekOfYear'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.Xor'>: <function bool_xor_sql>}
TYPE_MAPPING = {<Type.NCHAR: 'NCHAR'>: 'TEXT', <Type.NVARCHAR: 'NVARCHAR'>: 'TEXT', <Type.MEDIUMTEXT: 'MEDIUMTEXT'>: 'TEXT', <Type.LONGTEXT: 'LONGTEXT'>: 'TEXT', <Type.TINYTEXT: 'TINYTEXT'>: 'TEXT', <Type.MEDIUMBLOB: 'MEDIUMBLOB'>: 'BLOB', <Type.LONGBLOB: 'LONGBLOB'>: 'BLOB', <Type.TINYBLOB: 'TINYBLOB'>: 'BLOB', <Type.INET: 'INET'>: 'INET', <Type.BINARY: 'BINARY'>: 'BLOB', <Type.CHAR: 'CHAR'>: 'TEXT', <Type.FLOAT: 'FLOAT'>: 'REAL', <Type.UINT: 'UINT'>: 'UINTEGER', <Type.VARBINARY: 'VARBINARY'>: 'BLOB', <Type.VARCHAR: 'VARCHAR'>: 'TEXT', <Type.TIMESTAMP_S: 'TIMESTAMP_S'>: 'TIMESTAMP_S', <Type.TIMESTAMP_MS: 'TIMESTAMP_MS'>: 'TIMESTAMP_MS', <Type.TIMESTAMP_NS: 'TIMESTAMP_NS'>: 'TIMESTAMP_NS'}
STAR_MAPPING = {'except': 'EXCLUDE', 'replace': 'REPLACE'}
UNWRAPPED_INTERVAL_VALUES = (<class 'sqlglot.expressions.Column'>, <class 'sqlglot.expressions.Literal'>, <class 'sqlglot.expressions.Paren'>)
PROPERTIES_LOCATION = {<class 'sqlglot.expressions.AlgorithmProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.AutoIncrementProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.BlockCompressionProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.CharacterSetProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ChecksumProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.CollateProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.CopyGrantsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.Cluster'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ClusteredByProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DataBlocksizeProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.DefinerProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.DictRange'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DictProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DistKeyProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DistStyleProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.EngineProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ExecuteAsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ExternalProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.FallbackProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.FileFormatProperty'>: <Location.POST_WITH: 'POST_WITH'>, <class 'sqlglot.expressions.FreespaceProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.HeapProperty'>: <Location.POST_WITH: 'POST_WITH'>, <class 'sqlglot.expressions.InputModelProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.IsolatedLoadingProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.JournalProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.LanguageProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.LikeProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.LocationProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.LockingProperty'>: <Location.POST_ALIAS: 'POST_ALIAS'>, <class 'sqlglot.expressions.LogProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.MaterializedProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.MergeBlockRatioProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.NoPrimaryIndexProperty'>: <Location.POST_EXPRESSION: 'POST_EXPRESSION'>, <class 'sqlglot.expressions.OnProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.OnCommitProperty'>: <Location.POST_EXPRESSION: 'POST_EXPRESSION'>, <class 'sqlglot.expressions.Order'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.OutputModelProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.PartitionedByProperty'>: <Location.POST_WITH: 'POST_WITH'>, <class 'sqlglot.expressions.PartitionedOfProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.PrimaryKey'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.Property'>: <Location.POST_WITH: 'POST_WITH'>, <class 'sqlglot.expressions.RemoteWithConnectionModelProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ReturnsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.RowFormatProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.RowFormatDelimitedProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.RowFormatSerdeProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SampleProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SchemaCommentProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SerdeProperties'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.Set'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SettingsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SetProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.SortKeyProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SqlSecurityProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.StabilityProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.TemporaryProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.ToTableProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.TransientProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.TransformModelProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.MergeTreeTTL'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.VolatileProperty'>: <Location.UNSUPPORTED: 'UNSUPPORTED'>, <class 'sqlglot.expressions.WithDataProperty'>: <Location.POST_EXPRESSION: 'POST_EXPRESSION'>, <class 'sqlglot.expressions.WithJournalTableProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.WithSystemVersioningProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>}
def interval_sql(self, expression: sqlglot.expressions.Interval) -> str: View Source

    413        def interval_sql(self, expression: exp.Interval) -> str:
    414            multiplier: t.Optional[int] = None
    415            unit = expression.text("unit").lower()
    416
    417            if unit.startswith("week"):
    418                multiplier = 7
    419            if unit.startswith("quarter"):
    420                multiplier = 90
    421
    422            if multiplier:
    423                return f"({multiplier} * {super().interval_sql(exp.Interval(this=expression.this, unit=exp.var('day')))})"
    424
    425            return super().interval_sql(expression)

def tablesample_sql( self, expression: sqlglot.expressions.TableSample, seed_prefix: str = 'SEED', sep: str = ' AS ') -> str: View Source

    427        def tablesample_sql(
    428            self, expression: exp.TableSample, seed_prefix: str = "SEED", sep: str = " AS "
    429        ) -> str:
    430            return super().tablesample_sql(expression, seed_prefix="REPEATABLE", sep=sep)

SELECT_KINDS: Tuple[str, ...] = ()

##### Inherited Members

sqlglot.generator.Generator
  Generator
  NULL_ORDERING_SUPPORTED
  LOCKING_READS_SUPPORTED
  EXPLICIT_UNION
  WRAP_DERIVED_VALUES
  CREATE_FUNCTION_RETURN_AS
  MATCHED_BY_SOURCE
  SINGLE_STRING_INTERVAL
  INTERVAL_ALLOWS_PLURAL_FORM
  TABLESAMPLE_WITH_METHOD
  TABLESAMPLE_SIZE_IS_PERCENT
  LIMIT_ONLY_LITERALS
  GROUPINGS_SEP
  INDEX_ON
  QUERY_HINT_SEP
  IS_BOOL_ALLOWED
  DUPLICATE_KEY_UPDATE_WITH_SET
  LIMIT_IS_TOP
  RETURNING_END
  COLUMN_JOIN_MARKS_SUPPORTED
  EXTRACT_ALLOWS_QUOTES
  TZ_TO_WITH_TIME_ZONE
  VALUES_AS_TABLE
  ALTER_TABLE_INCLUDE_COLUMN_KEYWORD
  UNNEST_WITH_ORDINALITY
  AGGREGATE_FILTER_SUPPORTED
  COMPUTED_COLUMN_WITH_TYPE
  SUPPORTS_TABLE_COPY
  TABLESAMPLE_REQUIRES_PARENS
  COLLATE_IS_FUNC
  DATA_TYPE_SPECIFIERS_ALLOWED
  ENSURE_BOOLS
  CTE_RECURSIVE_KEYWORD_REQUIRED
  SUPPORTS_SINGLE_ARG_CONCAT
  TIME_PART_SINGULARS
  TOKEN_MAPPING
  PARAMETER_TOKEN
  RESERVED_KEYWORDS
  WITH_SEPARATED_COMMENTS
  EXCLUDE_COMMENTS
  EXPRESSIONS_WITHOUT_NESTED_CTES
  KEY_VALUE_DEFINITONS
  SENTINEL_LINE_BREAK
  pretty
  identify
  normalize
  pad
  unsupported_level
  max_unsupported
  leading_comma
  max_text_width
  comments
  dialect
  normalize_functions
  unsupported_messages
  generate
  preprocess
  unsupported
  sep
  seg
  pad_comment
  maybe_comment
  wrap
  no_identify
  normalize_func
  indent
  sql
  uncache_sql
  cache_sql
  characterset_sql
  column_sql
  columnposition_sql
  columndef_sql
  columnconstraint_sql
  computedcolumnconstraint_sql
  autoincrementcolumnconstraint_sql
  compresscolumnconstraint_sql
  generatedasidentitycolumnconstraint_sql
  generatedasrowcolumnconstraint_sql
  periodforsystemtimeconstraint_sql
  notnullcolumnconstraint_sql
  transformcolumnconstraint_sql
  primarykeycolumnconstraint_sql
  uniquecolumnconstraint_sql
  createable_sql
  create_sql
  clone_sql
  describe_sql
  prepend_ctes
  with_sql
  cte_sql
  tablealias_sql
  bitstring_sql
  hexstring_sql
  bytestring_sql
  rawstring_sql
  datatypeparam_sql
  datatype_sql
  directory_sql
  delete_sql
  drop_sql
  except_sql
  except_op
  fetch_sql
  filter_sql
  hint_sql
  index_sql
  identifier_sql
  inputoutputformat_sql
  national_sql
  partition_sql
  properties_sql
  root_properties
  properties
  with_properties
  locate_properties
  property_name
  property_sql
  likeproperty_sql
  fallbackproperty_sql
  journalproperty_sql
  freespaceproperty_sql
  checksumproperty_sql
  mergeblockratioproperty_sql
  datablocksizeproperty_sql
  blockcompressionproperty_sql
  isolatedloadingproperty_sql
  partitionboundspec_sql
  partitionedofproperty_sql
  lockingproperty_sql
  withdataproperty_sql
  withsystemversioningproperty_sql
  insert_sql
  intersect_sql
  intersect_op
  introducer_sql
  kill_sql
  pseudotype_sql
  objectidentifier_sql
  onconflict_sql
  returning_sql
  rowformatdelimitedproperty_sql
  withtablehint_sql
  indextablehint_sql
  historicaldata_sql
  table_sql
  pivot_sql
  version_sql
  tuple_sql
  update_sql
  values_sql
  var_sql
  into_sql
  from_sql
  group_sql
  having_sql
  connect_sql
  prior_sql
  join_sql
  lambda_sql
  lateral_sql
  limit_sql
  offset_sql
  setitem_sql
  set_sql
  pragma_sql
  lock_sql
  literal_sql
  escape_str
  loaddata_sql
  null_sql
  boolean_sql
  order_sql
  cluster_sql
  distribute_sql
  sort_sql
  ordered_sql
  matchrecognize_sql
  query_modifiers
  offset_limit_modifiers
  after_having_modifiers
  after_limit_modifiers
  select_sql
  schema_sql
  schema_columns_sql
  star_sql
  parameter_sql
  sessionparameter_sql
  placeholder_sql
  subquery_sql
  qualify_sql
  union_sql
  union_op
  unnest_sql
  where_sql
  window_sql
  partition_by_sql
  windowspec_sql
  withingroup_sql
  between_sql
  bracket_sql
  all_sql
  any_sql
  exists_sql
  case_sql
  constraint_sql
  nextvaluefor_sql
  extract_sql
  trim_sql
  convert_concat_args
  concat_sql
  concatws_sql
  check_sql
  foreignkey_sql
  primarykey_sql
  if_sql
  matchagainst_sql
  jsonkeyvalue_sql
  formatjson_sql
  jsonobject_sql
  jsonarray_sql
  jsonarrayagg_sql
  jsoncolumndef_sql
  jsonschema_sql
  jsontable_sql
  openjsoncolumndef_sql
  openjson_sql
  in_sql
  in_unnest_op
  return_sql
  reference_sql
  anonymous_sql
  paren_sql
  neg_sql
  not_sql
  alias_sql
  aliases_sql
  attimezone_sql
  add_sql
  and_sql
  xor_sql
  connector_sql
  bitwiseand_sql
  bitwiseleftshift_sql
  bitwisenot_sql
  bitwiseor_sql
  bitwiserightshift_sql
  bitwisexor_sql
  cast_sql
  currentdate_sql
  collate_sql
  command_sql
  comment_sql
  mergetreettlaction_sql
  mergetreettl_sql
  transaction_sql
  commit_sql
  rollback_sql
  altercolumn_sql
  renametable_sql
  altertable_sql
  add_column_sql
  droppartition_sql
  addconstraint_sql
  distinct_sql
  ignorenulls_sql
  respectnulls_sql
  intdiv_sql
  dpipe_sql
  div_sql
  overlaps_sql
  distance_sql
  dot_sql
  eq_sql
  propertyeq_sql
  escape_sql
  glob_sql
  gt_sql
  gte_sql
  ilike_sql
  ilikeany_sql
  is_sql
  like_sql
  likeany_sql
  similarto_sql
  lt_sql
  lte_sql
  mod_sql
  mul_sql
  neq_sql
  nullsafeeq_sql
  nullsafeneq_sql
  or_sql
  slice_sql
  sub_sql
  trycast_sql
  log_sql
  use_sql
  binary
  function_fallback_sql
  func
  format_args
  text_width
  format_time
  expressions
  op_expressions
  naked_property
  set_operation
  tag_sql
  token_sql
  userdefinedfunction_sql
  joinhint_sql
  kwarg_sql
  when_sql
  merge_sql
  tochar_sql
  dictproperty_sql
  dictrange_sql
  dictsubproperty_sql
  oncluster_sql
  clusteredbyproperty_sql
  anyvalue_sql
  querytransform_sql
  indexconstraintoption_sql
  indexcolumnconstraint_sql
  nvl2_sql
  comprehension_sql
  columnprefix_sql
  opclass_sql
  predict_sql
  forin_sql
  refresh_sql
  operator_sql